import numpy as np
import pandas as pd
from datetime import datetime
import matplotlib.pyplot as plt

###############################################################################

def create_df_with_dates (array, 
                          start = '2020-01-01 00:00:00', 
                          freq = '1min'):
    """
    Parameters
    ----------
    array: pandas.DataFrame or numpy.array
        Original data.
    start: string, optional
        Start timestamp.
    freq: string, optional
        Sampling interval.
    Returns
    ----------                
    df: pandas.DataFrame
        Processed data.
    """    
    df = pd.DataFrame(array)
    df.index = pd.date_range(start = start, 
                             periods = df.shape[0],
                             freq=freq)
    return df

###############################################################################

def create_df_with_noise (array, 
                          noise_frac, 
                          max_index_for_noise):
    """
    Adds artificial measurement noise to data.
    
    Parameters
    ----------
    array: pandas.DataFrame or numpy.array
        Original data.
    noise_frac: float 
        Fraction (between 0 and 1) of the total amplitude of the variable
        that will be used as the noise standard deviation.
    max_index_for_noise: int
        Maximum index to consider the amplitude 
        in the standard deviation calculation.
    Returns
    ----------                
    df: pandas.DataFrame
        Processed data.
    """    

    df = pd.DataFrame(array)

    sigma_noise = noise_frac*(np.amax(array[:max_index_for_noise,:], 
                                      axis=0)-
                              np.amin(array[:max_index_for_noise,:], 
                                      axis=0))

    for i in range(df.shape[1]):
        df.iloc[:,i] = (df.iloc[:,i] + 
                        sigma_noise[i]*np.random.randn(df.shape[0]))
        
    return df
    
###############################################################################

def align_dfs_by_rows (df1, df2):
    """
    Aligns DataFrames by rows.

    Parameters
    ----------
    df1, df2: pandas.DataFrame
        Original data.
    Returns
    ----------                
    new_df1, new_df2: pandas.DataFrame
        Processed data.
    """
    
    new_df1 = df1.loc[df1.index.isin(df2.index)]
    new_df2 = df2.loc[df2.index.isin(df1.index)]

    return new_df1, new_df2

###############################################################################

def spearmanr_dendrogram(df, figsize = (18,8)):
    """
    Generates a dendrogram of Spearman correlations.
    
    Parameters
    ----------
    df: pandas.DataFrame
        Dados to be analyzed.
    figsize: tuple of ints, optional
        Figure dimensions.
    """   
    import scipy.cluster.hierarchy
    import scipy.stats
    
    df = pd.DataFrame(df)
    
    corr = np.round(scipy.stats.spearmanr(df).correlation, 4)
    corr_condensed = scipy.cluster.hierarchy.distance.squareform(1-corr)
    z = scipy.cluster.hierarchy.linkage(corr_condensed, method='average')
    plt.figure(figsize=figsize)
    scipy.cluster.hierarchy.dendrogram(z, labels = df.columns.tolist(), 
                                       orientation='left', leaf_font_size=16)
    plt.xlabel('Spearman Correlation')
    plt.ylabel('Variable clusters')
    plt.show()

###############################################################################

def train_val_test_split (data, start_train, end_train, 
                          end_validation, end_test, 
                          tags_X = None, tags_Y = None):
    """
    Separates the data into consecutive portions of 
    train, validation, and test, returning 3 DataFrames.
    It can also separate into predictor variables (X) and 
    predicted variables (Y), which in this case will return 6 DataFrames.
        
    Parameters
    ----------
    data: pandas.DataFrame
        Data to be separated.
    start_train: string
        Start timestamp of the train portion.
    end_train: string
        End timestamp of the train portion.
    end_validation: string
        End timestamp of the validation portion.
    end_test: string
        End timestamp of the test portion.
    tags_X: list of strings
        Variables to be considered in the X set.
    tags_Y: list of strings
        Variables to be considered in the Y set.
    Returns
    ----------                
    : pandas.DataFrames
        Separated data.
    """               
    train_data = data.loc[start_train:end_train]
    validation_data = data.loc[end_train:end_validation].iloc[1:,:]
    test_data = data.loc[end_validation:end_test].iloc[1:,:]

    if tags_Y is not None:

        if not isinstance(tags_Y, list): tags_Y = [tags_Y]
        
        train_data_Y = train_data.loc[:,tags_Y]
        validation_data_Y = validation_data.loc[:,tags_Y]
        test_data_Y = test_data.loc[:,tags_Y]

        if tags_X is not None:

            if not isinstance(tags_X, list): tags_X = [tags_X]
            
            train_data_X = train_data.loc[:,tags_X]
            validation_data_X = validation_data.loc[:,tags_X]
            test_data_X = test_data.loc[:,tags_X]

        else:

            dif = train_data.columns.difference(train_data_Y.columns)
            train_data_X = train_data[dif]
            validation_data_X = validation_data[dif]
            test_data_X = test_data[dif]
            
        return (train_data_X, validation_data_X, test_data_X, 
                train_data_Y, validation_data_Y, test_data_Y)
            
    else:
        
        if tags_X is not None:

            if not isinstance(tags_X, list): tags_Y = [tags_X]
            
            train_data = train_data.loc[:,tags_X]
            validation_data = validation_data.loc[:,tags_X]
            test_data = test_data.loc[:,tags_X]           
            
        return (train_data, validation_data, test_data)
            
###############################################################################

def complete_analysis (model, X_train, X_validation, X_test, 
                       Y_train = None , Y_validation = None, Y_test = None,
                       lim_conf = 0.99,
                       f_pp_train = ['remove_empty_variables',
                                     'ffill_nan',
                                     'remove_frozen_variables',
                                     'normalize'],
                       a_pp_train = None,
                       f_pp_test = ['replace_nan_with_values',
                                    'normalize'],
                       a_pp_test = None,
                       logy = True, 
                       metrics = None, 
                       X_pred_to_plot = None,
                       count_limit = 1,
                       count_window_size = 0,
                       fault_start = None,
                       fault_end = None):
    """
    Performs a complete monitoring analysis, with train, validation, and test.

    Parameters
    ----------
    model: BibMon model
        Model to be considered in the analysis.
    X_train: pandas.DataFrame or numpy.array
        Training data X.
    X_validation: pandas.DataFrame or numpy.array
        Validation data X.
    X_test: pandas.DataFrame or numpy.array
        Test data X.
    Y_train: pandas.DataFrame or numpy.array, optional
        Training data Y.
    Y_validation: pandas.DataFrame or numpy.array, optional
        Validation data Y.
    Y_test: pandas.DataFrame or numpy.array, optional
        Test data Y.
    lim_conf: float, optional
        Confidence limit for the detection index.
    f_pp_train: list, optional
        List containing strings with names of functions to be used 
        in pre-processing the train data (the functions are defined in the 
        PreProcess class, in the BibMon_Tools.py file).
    a_pp_train: dict, optional
        Dictionary containing the parameters to be provided
        to each function to perform pre-processing of the train data, in
        the format {'functionname__argname': argvalue, ...}.
    f_pp_test: list, optional
        List containing strings with names of functions to be used 
        in pre-processing the test data (the functions are defined in the 
        PreProcess class, in the BibMon_Tools.py file).
    a_pp_test: dict, optional
        Dictionary containing the parameters to be provided
        to each function to perform pre-processing of the test data, in
        the format {'functionname__argname': argvalue, ...}.
    logy: boolean, optional
        If use logarithmic scale in the SPE plots.
    metrics: list of functions, optional
        Functions for calculating metrics to be displayed in the title of 
        the graph.
    X_pred_to_plot: string, optional
        In case the model is a reconstruction model (i.e., self.has_Y = False),
        indicates which column of X to plot along with the prediction.
    count_limit: int, optional
        Limit of points to be considered in the window 
        for the count alarm to sound.
    count_window_size: int, optional
        Window sizes used in count alarm calculation. 
    fault_start: string, optional
        Start timestamp of the fault.
    fault_end: string, optional
        End timestamp of the fault.
    """               
    fig, ax = plt.subplots(3,2, figsize = (15,12))

    cond_to_plot_pred = (model.has_Y or 
                        ((not model.has_Y) and (X_pred_to_plot is not None)))

    ######## TRAINING ########

    model.fit(X_train, Y_train, 
              f_pp = f_pp_train,
              a_pp = a_pp_train,
              f_pp_test = f_pp_test,
              a_pp_test = a_pp_test,
              lim_conf = lim_conf,
              redefine_limit = False)
                                
    # PLOTTING SPE
                
    model.plot_SPE(ax = ax[0,0], logy = logy)
    ax[0,0].set_title('Training')

    # PLOTTING PREDICTIONS

    if cond_to_plot_pred:
        model.plot_predictions(ax = ax[0,1], train_or_test = 'train', 
                               X_pred_to_plot = X_pred_to_plot,
                               metrics = metrics)

    ######## VALIDATION ########

    model.predict(X_validation, Y_validation, 
                  count_window_size = count_window_size, 
                  redefine_limit = True)

    # PLOTTING SPE

    model.plot_SPE(ax = ax[1,0], train_or_test = 'test', logy = logy)
    ax[1,0].set_title('Validation')

    # PLOTTING PREDICTIONS

    if cond_to_plot_pred:
        model.plot_predictions(ax = ax[1,1], train_or_test = 'test', 
                               X_pred_to_plot = X_pred_to_plot, 
                               metrics = metrics)
        
    ######## TEST ########
        
    model.predict(X_test, Y_test, 
                  count_window_size = count_window_size, 
                  count_limit = count_limit,
                  redefine_limit = False)

    # PLOTTING SPE

    model.plot_SPE(ax = ax[2,0], train_or_test = 'test', logy = logy)
    ax[2,0].set_title('Test')

    if fault_start is not None:
        ax[2,0].axvline(datetime.strptime(str(fault_start),
                                          '%Y-%m-%d %H:%M:%S'), ls = '--')
    if fault_end is not None:
        ax[2,0].axvline(datetime.strptime(str(fault_end),
                                          '%Y-%m-%d %H:%M:%S'), ls = '--')

    # PLOTTING PREDICTIONS

    if cond_to_plot_pred:
        model.plot_predictions(ax = ax[2,1], train_or_test = 'test', 
                               X_pred_to_plot = X_pred_to_plot,
                               metrics = metrics)

        if fault_start is not None:
            ax[2,1].axvline(datetime.strptime(str(fault_start),
                                              '%Y-%m-%d %H:%M:%S'), ls = '--')
        if fault_end is not None:
            ax[2,1].axvline(datetime.strptime(str(fault_end),
                                              '%Y-%m-%d %H:%M:%S'), ls = '--')
        
    fig.tight_layout();
            
##############################################################################

def comparative_table (models, X_train, X_validation, X_test, 
                       Y_train = None , Y_validation = None, Y_test = None,
                       lim_conf = 0.99,
                       f_pp_train = ['remove_empty_variables',
                                     'ffill_nan',
                                     'remove_frozen_variables',
                                     'normalize'],
                       a_pp_train = None,
                       f_pp_test = ['replace_nan_with_values',
                                    'normalize'],
                       a_pp_test = None,
                       logy = True, metrics = None,
                       X_pred_to_plot = None,
                       count_limit = 1,
                       count_window_size = 0,
                       fault_start = None,
                       fault_end = None,
                       mask = None,
                       times = True,
                       plot_SPE = True,
                       plot_predictions = True,
                       fit_model = True):

    """
    Performs complete monitoring analysis of multiple models and builds
    comparative result tables.

    Parameters
    ----------
    models: list of BibMon models
        Models to be considered in the analysis.
    X_train: pandas.DataFrame or numpy.array
        Training data X.
    X_validation: pandas.DataFrame or numpy.array
        Validation data X.
    X_test: pandas.DataFrame or numpy.array
        Test data X.
    Y_train: pandas.DataFrame or numpy.array, optional
        Training data Y.
    Y_validation: pandas.DataFrame or numpy.array, optional
        Validation data Y.
    Y_test: pandas.DataFrame or numpy.array, optional
        Test data Y.
    lim_conf: float, optional
        Confidence limit for the detection index.
    f_pp_train: list, optional
        List containing strings with names of functions to be used 
        in pre-processing the train data (the functions are defined in the 
        PreProcess class, in the BibMon_Tools.py file).
    a_pp_train: dict, optional
        Dictionary containing the parameters to be provided
        to each function to perform pre-processing of the train data, in
        the format {'functionname__argname': argvalue, ...}.
    f_pp_test: list, optional
        List containing strings with names of functions to be used 
        in pre-processing the test data (the functions are defined in the 
        PreProcess class, in the BibMon_Tools.py file).
    a_pp_test: dict, optional
        Dictionary containing the parameters to be provided
        to each function to perform pre-processing of the test data, in
        the format {'functionname__argname': argvalue, ...}.
    logy: boolean, optional
        If use logarithmic scale in the SPE plots.
    metrics: list of functions, optional
        Functions for calculating metrics to be displayed in the title of 
        the graph.
    X_pred_to_plot: string, optional
        In case the model is a reconstruction model (i.e., self.has_Y = False),
        indicates which column of X to plot along with the prediction.
    count_limit: int, optional
        Limit of points to be considered in the window 
        for the count alarm to sound.
    count_window_size: int, optional
        Window sizes used in count alarm calculation.
    fault_start: string, optional
        Start timestamp of the fault.
    fault_end: string, optional
        End timestamp of the fault.
    mask: numpy.array, optional
        Boolean array indicating the indices where the process is
        in fault.
    times: boolean, optional
        If execution times should be calculated.
    plot_SPE: boolean, optional
        If SPE plots should be plotted.
    plot_predictions: boolean, optional
        If prediction plots should be plotted.
    fit_model: boolean, optional
        If models should be trained.
    Returns
    ----------
    : list of pandas.DataFrames
        List with the generated tables (prediction and/or detection).
    """
    
    n = len(models)

    if plot_SPE:
        fig_spe, ax_spe = plt.subplots(n, 3, figsize = (15, 4*n))

    if plot_predictions:
        fig_pred, ax_pred = plt.subplots(n, 3, figsize = (15, 4*n))
        
    train_metrics = {}     
    validation_metrics = {}        
    test_metrics = {}        

    detection_alarms = {}
    false_detection_rates = {}
    false_alarm_rates = {}

    for i in range(len(models)):
        
        model = models[i]
        
        cond_to_plot_pred = (model.has_Y or 
                            ((not model.has_Y) and 
                            (X_pred_to_plot is not None)))

        ######## TRAINING ########
        
        if fit_model:
            model.fit(X_train, Y_train, 
                      f_pp = f_pp_train,
                      a_pp = a_pp_train, 
                      f_pp_test = f_pp_test,
                      a_pp_test = a_pp_test,
                      lim_conf = lim_conf,
                      redefine_limit = False)
        
        # TERMS FOR PREDICTION TABLE
        
        if metrics is not None:
            if model.has_Y:
                true = model.Y_train_orig
                pred = model.Y_train_pred_orig
            else:
                if X_pred_to_plot is not None:
                    true = model.X_train_orig[X_pred_to_plot]
                    pred = model.X_train_pred_orig[X_pred_to_plot]                
            for mr in metrics:
                true, pred = align_dfs_by_rows(true.dropna(), pred)
                train_metrics[model.name+': '+mr.__name__] = mr(true, pred)
                                    
        # PLOTTING SPE
                
        if plot_SPE:
            model.plot_SPE(ax = ax_spe[i,0], logy = logy)
            ax_spe[i,0].set_title('Training')

        # PLOTTING PREDICTIONS
            
        if plot_predictions:
            if cond_to_plot_pred:
                model.plot_predictions(ax = ax_pred[i,0], 
                                        train_or_test = 'train', 
                                        X_pred_to_plot = X_pred_to_plot,
                                        metrics = metrics)

        ######## VALIDATION ########
                
        model.predict(X_validation, Y_validation, 
                      count_window_size = count_window_size, 
                      redefine_limit = True)

        # TERMS FOR PREDICTION TABLE
        
        if metrics is not None:
            if model.has_Y:
                true = model.Y_test_orig
                pred = model.Y_test_pred_orig
            else:
                if X_pred_to_plot is not None:
                    true = model.X_test_orig[X_pred_to_plot]
                    pred = model.X_test_pred_orig[X_pred_to_plot]                
            for mr in metrics:
                true, pred = align_dfs_by_rows(true.dropna(), pred)
                validation_metrics[model.name+': '+mr.__name__]= mr(true, pred)

        # PLOTTING SPE
                
        if plot_SPE:
            model.plot_SPE(ax = ax_spe[i,1], train_or_test = 'test', 
                           logy=logy)
            ax_spe[i,1].set_title(model.name+'\n\nValidation')

        # PLOTTING PREDICTIONS
            
        if plot_predictions:
            if cond_to_plot_pred:
                model.plot_predictions(ax = ax_pred[i,1], 
                                       train_or_test = 'test', 
                                       X_pred_to_plot = X_pred_to_plot, 
                                       metrics = metrics)
                ax_pred[i,1].set_title(model.name+'\n\n'+
                                       ax_pred[i,1].get_title())
                
        ######## TEST ########
                
        model.predict(X_test, Y_test, 
                      count_window_size = count_window_size, 
                      count_limit = count_limit,
                      redefine_limit = False)

        # TERMS FOR PREDICTION TABLE
        
        if metrics is not None:
            if model.has_Y:
                true = model.Y_test_orig
                pred = model.Y_test_pred_orig
            else:
                if X_pred_to_plot is not None:
                    true = model.X_test_orig[X_pred_to_plot]
                    pred = model.X_test_pred_orig[X_pred_to_plot]                
            for mr in metrics:
                true, pred = align_dfs_by_rows(true.dropna(), pred)
                test_metrics[model.name+': '+mr.__name__] =  mr(true, pred)

        # TERMS FOR DETECTION TABLE
        
        if mask is None:
            if fault_start is not None:            
                for key, value in model.alarms.items():
                    if fault_end is not None:
                        detection_alarms[model.name+': '+key] = \
                            value[fault_start:fault_end][:-1].mean()
                        false_detection_rates[model.name+': '+key] = \
                            pd.concat([value[:fault_start][:-1], 
                                    value[fault_end:]]).mean()
                    else:
                        detection_alarms[model.name+': '+key] = \
                            value[fault_start:fault_end].mean()
                        false_detection_rates[model.name+': '+key] = \
                            value[:fault_start][:-1].mean()
        else:
            for key, value in model.alarms.items():
                detection_alarms[model.name+': '+key] = \
                    mask[mask==1].eq(value[value==1]).astype(int).mean()
                false_detection_rates[model.name+': '+key] = \
                    1-mask[mask==0].eq(value[value==0]).astype(int).mean()

        # PLOTTING SPE
                
        if plot_SPE:

            model.plot_SPE(ax = ax_spe[i,2], train_or_test='test',
                           logy = logy)
            ax_spe[i,2].set_title('Test')
            
            if fault_start is not None:
                ax_spe[i,2].axvline(datetime.strptime(str(fault_start),
                                                      '%Y-%m-%d %H:%M:%S'), 
                                                      ls='--')
            if fault_end is not None:
                ax_spe[i,2].axvline(datetime.strptime(str(fault_end),
                                                      '%Y-%m-%d %H:%M:%S'), 
                                                      ls='--')

        # PLOTTING PREDICTIONS
                
        if plot_predictions:
            if cond_to_plot_pred:
                model.plot_predictions(ax = ax_pred[i,2], 
                                        train_or_test = 'test', 
                                        X_pred_to_plot = X_pred_to_plot,
                                        metrics = metrics)
                if fault_start is not None:
                    ax_pred[i,2].axvline(datetime.strptime(str(fault_start),
                                                 '%Y-%m-%d %H:%M:%S'), ls='--')
                if fault_end is not None:
                    ax_pred[i,2].axvline(datetime.strptime(str(fault_end),
                                                 '%Y-%m-%d %H:%M:%S'), ls='--')

        if plot_SPE:            
            fig_spe.tight_layout();
        if plot_predictions:
            fig_pred.tight_layout();
            
    ######## GENERATING FINAL TABLES ########
        
    return_tables = []
        
    # PREDICTION

    if metrics is not None:

        prediction_table = pd.DataFrame([train_metrics, validation_metrics,
                                         test_metrics], 
                                        index = ['Train',
                                                 'Validation',
                                                 'Test']).T
                                                            
        models_names = [models[i].name for i in range(len(models))]
        metrics_names = [metrics[i].__name__ for i in range(len(metrics))]
        
        iterables = [models_names, metrics_names]

        index = pd.MultiIndex.from_product(iterables, 
                                           names = ['Models', 'Metrics'])
        
        message = ("The MultiIndex table is not of the right size. "+
                   "This could happen when two models have the same name.")
        
        try:
            prediction_table.index = index
        except ValueError as err:
            raise ValueError(message) from err
        
        return_tables.append(prediction_table.swaplevel().sort_index(axis=0))

    # DETECTION     
        
    if fault_start is not None:
        
        detection_table = pd.DataFrame([detection_alarms, 
                                        false_detection_rates],
                                       index = ['FDR','FAR']).T
        
        models_names = [models[i].name for i in range(len(models))]
        alarms_names = [list(model.alarms.keys())[i] for i in 
                range(len(list(model.alarms.keys())))]  
        
        iterables = [models_names, alarms_names]

        index = pd.MultiIndex.from_product(iterables, 
                                        names = ['Models', 'Alarms'])
        
        detection_table.index = index

        return_tables.append(detection_table.swaplevel().sort_index(axis=0))
        
    # COMPUTATIONAL TIMES
        
    if times:
        
        times_dict = {}
        
        times_dict['Train'] = [1e6*models[i].train_time/X_train.shape[0]
                        for i in range(len(models))]
        times_dict['Test'] = [1e6*models[i].test_time/X_test.shape[0] 
                        for i in range(len(models))]
        
        times_df = pd.DataFrame(times_dict, 
                        index=[models[i].name for i in range(len(models))])
        
        return_tables.append(times_df)
        
    return return_tables


from ._autoencoder import Autoencoder
from ._pca import PCA
from ._esn import ESN
from ._sbm import SBM
from ._sklearn_regressor import sklearnRegressor
from ._preprocess import PreProcess
from ._load_data import load_tennessee_eastman, load_real_data
from ._bibmon_tools import train_val_test_split, complete_analysis, comparative_table, spearmanr_dendrogram, create_df_with_dates, create_df_with_noise, align_dfs_by_rows

__all__ = ['Autoencoder','PCA','ESN','SBM',
	   'sklearnRegressor', 'PreProcess',
           'load_tennessee_eastman', 'load_real_data', 
           'train_val_test_split', 'complete_analysis', 'comparative_table',
	       'spearmanr_dendrogram', 'create_df_with_dates',
           'create_df_with_noise', 'align_dfs_by_rows']



import os
import pandas as pd
import importlib.resources as pkg_resources

from ._bibmon_tools import create_df_with_dates
from . import real_process_data, tennessee_eastman 

###############################################################################

def load_tennessee_eastman (train_id = 0, test_id = 0):
    """
    Load the 'Tennessee Eastman Process' benchmark data.

    Parameters
    ----------
    train_id: int, optional
        Identifier of the training data.
        No fault: 0. With faults: 1 to 20.
    test_id: int, optional
        Identifier of the test data.
        No fault: 0. With faults: 1 to 20.
    Returns
    ----------                
    train_df: pandas.DataFrame
        Training data.
    test_df: pandas.DataFrame
        Test data.
    """    
        
    tags1 = ['XMEAS('+str(ii)+')' for ii in range(1,42)]
    tags2 = ['XMV('+str(ii)+')' for ii in range(1,12)]
    tags = tags1 + tags2
    
    file_train = f'd{train_id}.dat'
    file_test = f'd{test_id}_te.dat'

    if len(file_train) == 6:
        file_train = file_train[:2]+'0'+file_train[2:]
        
    if len(file_test) == 9:
        file_test = file_test[:1]+'0'+file_test[1:]

    with pkg_resources.path(tennessee_eastman, file_train) as filepath:

        if file_train == 'd00.dat':
            
            tmp1 = pd.read_csv(filepath,sep='\t',
                               names=['0'])
            tmp2 = pd.DataFrame([tmp1.T.iloc[0,i].strip() for 
                                i in range(tmp1.shape[0])])
            train_df = pd.DataFrame()

            for ii in range(52):
                train_df[tags[ii]]=[float(s) for s in tmp2[0][ii].split('  ')]
                
            train_df = create_df_with_dates(train_df, 
                                            freq = '3min')
            
        else:

            train_df = create_df_with_dates(filepath, sep='\s+',names=tags,
                                            freq = '3min')

    with pkg_resources.path(tennessee_eastman, file_test) as filepath:

        test_df = create_df_with_dates(pd.read_csv(filepath, 
                                                   sep = '\s+', 
                                                   names = tags),
                                       start = '2020-02-01 00:00:00',
                                       freq = '3min')

    return train_df, test_df

###############################################################################

def load_real_data ():
    """
    Load a sample of real process data.
    The variables have been anonymized for availability in the library.
    
    Returns
    ----------                
    : pandas.DataFrame
        Process data.
    """    

    with pkg_resources.path(real_process_data,'real_process_data.csv') as file:
        return pd.read_csv(file,index_col = 0, parse_dates = True)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Sep  4 00:04:38 2020

@author: afranio
"""

import pandas as pd
import bibmon

#%%
def test_preprocess():

    # load data
    data = bibmon.load_real_data()
    data = data.apply(pd.to_numeric, errors='coerce')
    
    # define functions to be tested
    
    pp = bibmon.PreProcess()
    
    funcs = [f for f in dir(pp) if (f[0] != '_' and callable(getattr(pp,f)))]
    funcs = [f for f in funcs if f not in ['apply','back_to_units']]
    
    # define training and test sets
        
    (X_train, _, 
     X_test, Y_train, 
     _, Y_test) = bibmon.train_val_test_split(data, 
                                           start_train = '2017-12-24T12:00', 
                                           end_train = '2018-01-01T00:00', 
                                           end_validation = '2018-01-02T00:00', 
                                           end_test = '2018-01-04T00:00',
                                           tags_Y = 'tag100')
                                                         
    for f in funcs:
        
        print(f)
        
        ppX = bibmon.PreProcess(f_pp = [f])
        ppY = bibmon.PreProcess(f_pp = [f], is_Y = True)

        X_train_proc = ppX.apply(X_train, train_or_test = 'train')
        Y_train_proc = ppY.apply(Y_train, train_or_test = 'train')

        assert(isinstance(X_train_proc, pd.DataFrame))
        assert(isinstance(Y_train_proc, pd.DataFrame))
        assert(Y_train_proc.columns == Y_train.columns)
        
        X_test_proc = ppX.apply(X_test, train_or_test = 'test')
        Y_test_proc = ppY.apply(Y_test, train_or_test = 'test')
      
        assert(isinstance(X_test_proc, pd.DataFrame))
        assert(isinstance(Y_test_proc, pd.DataFrame))
        assert(Y_test_proc.columns == Y_test.columns)
        
        if 'nan' in f:
            assert(X_train_proc.isnull().sum().sum()==0)
            assert(Y_train_proc.isnull().sum().sum()==0)
            assert(X_test_proc.isnull().sum().sum()==0)
            assert(Y_test_proc.isnull().sum().sum()==0)


import copy
import time
import optuna
import numpy as np
import pandas as pd
import seaborn as sns
import sklearn.model_selection
import matplotlib.pyplot as plt

from abc import ABC, abstractmethod

from ._alarms import detecOutlier
from ._preprocess import PreProcess
from ._bibmon_tools import align_dfs_by_rows

###############################################################################

class GenericModel (ABC):
    """
    Abstract class used as a base for creating child classes with
    specific models (PCA, sklearnRegressor, etc).

    Attributes
    ----------
            
        * Data sets:
        ----------
        X_train, X_test, Y_train, Y_test: pandas.DataFrame
            Normalized training and testing data sets.
            X refers to the input and Y to the output (when there is Y).
            
        X_train_orig, X_test_orig, 
        Y_train_orig, Y_test_orig: pandas.DataFrame
            Original training and testing data sets (before pre-processing). 
            X refers to the input and Y to the output (when there is Y).       
            
        * Predictions:
        ----------
        X_train_pred, X_test_pred 
        (or Y_train_pred, Y_test_pred): pandas.DataFrame
            Reconstructions (or predictions) 
            provided by the model in training and testing.
            
        X_train_pred_orig, X_test_pred_orig
        (or Y_train_pred_orig, Y_test_pred_orig): pandas.DataFrame
            Reconstructions (or predictions) 
            provided by the model in training and testing
            in the original units.
        
        * Detection, contribution and prediction metrics:
        ----------
        SPE_train, SPE_test: pd.Series
            Square Prediction Error. One-dimensional vectors containing
            the SPE values calculated at each time step in 
            training and testing.  
                
        * Training parameters:
        ----------
        Mux, SDx, Muy, SDy: pandas.Series
            Means and standard deviations of the variables 
            in the training period.
            
        limSPE_train, SPE_mean: float
            Limit for detection and mean of the SPE obtained in training.

        limSPE: float
            Potentially redefined detection limit in the validation stage.
            
        lim_conf: float
            Confidence limit for the detection index.

        train_time: float
            Computational training time.
            
        index_train: pandas.DatetimeIndex
            Indexes of the training set.
            
        * Testing parameters:
        ----------            
        count_window_size: int
            Window sizes used in alarm calculations.
        
        outlier_alarm,
        count_alarm: numpy.array
            One-dimensional arrays that store the various alarms for each
            time step.
            
        alarms: dict
            Dictionary containing the mentioned alarms above.
        
        count_limit: int
            Limit of points to be considered for the count_alarm to trigger.

        test_time: float
            Computational testing time.

        has_alarm_windows: boolean
            Indicator of testing step with or without windows for the alarms.
            
        * Other attributes:
        ----------
        preproc_X, preproc_Y: PreProcess
            Object of the PreProcess class (defined in the _bibmon_tools.py 
            file) that contains the preprocessing functions and arguments
            to be applied in training and testing.
        tags_X, tags_Y: list of str
            tags kept in the data set, obtained in pre-training.
        test_tags_X, test_tags_Y: list of str
            tags used in testing.
        has_Y: boolean
            indicator of the existence or absence of 
            a predicted dataset (Y) in the model.
            
    Methods
    ----------
        train_core (@abstractmethod)
        map_from_X (@abstractmethod)
        set_hyperparameters
        load_model
        pre_train
        train
        pre_test
        test
        plot_SPE
        fit
        predict       
    """

    ###########################################################################
    
    @abstractmethod
    def train_core (self):
        """
        The core of the training algorithm, that is,
        all the necessary steps between `pre_train()` and
        the calculation of the prediction or reconstruction 
        in training by `map_from_X()`.
        """
        pass

    ###########################################################################
            
    @abstractmethod
    def map_from_X (self, X):
        """
        Receives a data matrix X and returns a matrix of predicted 
        or reconstructed values.

        Parameters
        ----------
        X: numpy.array
            Window X of data for prediction or reconstruction.  
            
        Returns
        ----------                
        : numpy.array
        Reconstructed X (or predicted Y).
        """  
        pass

    ###########################################################################
            
    def set_hyperparameters (self, params_dict):
        """
        Receives a dict with hyperparameters to be assigned in the model.

        Parameters
        ----------
        params_dict: dict
            Dictionary with hyperparameter values

        """          
        for key, value in params_dict.items():
            setattr(self, key, value)

    ###########################################################################
        
    def load_model (self, limSPE, SPE_mean, count_window_size,
                    Mux, SDx, Muy = None, SDy = None):
        """
        Receives parameters from a previously trained model for
        making predictions and tests without the need for training.

        Parameters
        ----------                     
        limSPE: float
            Detection limit and mean of the SPE.
        SPE_mean: float
            Mean of the SPE.
        count_window_size: int
            Window sizes used in count alarms calculation.
        Mux: pandas.Series
            Means of the X variables in the training period.
        SDx: pandas.Series
            Standard deviations of the X variables in
            the training period.
        Muy: pandas.Series, optional
            Means of the Y variables in the training period.
        SDy: pandas.Series, optional
            Standard deviations of the Y variables in the training period.    
        """  
    
        self.MuX = pd.Series(Mux)
        self.SDX = pd.Series(SDx)
        self.MuY = pd.Series(Muy)
        self.SDY = pd.Series(SDy)
        self.limSPE = limSPE
        self.SPE_mean = SPE_mean
        self.count_window_size = count_window_size

    ###########################################################################

    def pre_train (self, X_train, Y_train = None, 
                   f_pp = ['remove_empty_variables',
                           'ffill_nan',
                           'remove_frozen_variables',
                           'normalize'], 
                   a_pp = None,
                   f_pp_test = ['replace_nan_with_values',
                                'normalize'], 
                   a_pp_test = None):
        """
        Receives the data for model training and prepares them for training.

        Parameters
        ----------
        X_train: pandas.DataFrame or numpy.ndarray
            Window of X data used in training.
        Y_train: pandas.DataFrame or numpy.ndarray, optional
            Window of Y data used in training.
        f_pp: list, optional
            List containing strings with names of functions to be used 
            in pre-processing the training data (the functions are defined
            in the PreProcess class, in the BibMon_Tools.py file)
        f_pp_test: list, optional
            List containing strings with names of functions to be used 
            in pre-processing the testing data (the functions are defined 
            in the PreProcess class, in the BibMon_Tools.py file)
        a_pp: dict, optional
            Dictionary containing the parameters to be provided
            to each function to perform pre-processing of the training data, 
            in the format {'functionname__argname': argvalue, ...}
        a_pp_test: dict, optional
            Dictionary containing the parameters to be provided
            to each function to perform pre-processing of the testing data, ]
            in the format {'functionname__argname': argvalue, ...}
        """
                
        # storing complete training set
        
        self.X_train = pd.DataFrame(X_train)
        self.Y_train = pd.DataFrame(Y_train)

        # data set that will keep the original units

        self.X_train_orig = copy.deepcopy(self.X_train)
        if self.has_Y:
            self.Y_train_orig = copy.deepcopy(self.Y_train)

        # defining the training medians as the default values 
        # to replace NaNs during testing
        
        if f_pp_test is not None:
            if 'replace_nan_with_values' in f_pp_test:
                if a_pp_test is None:
                    a_pp_test = {}
                a_pp_test['replace_nan_with_values__val'] = \
                    self.X_train.median()

        # applying pre-processing functions

        if f_pp is not None:
            self.preproc_X = PreProcess(f_pp, a_pp)
            self.X_train = self.preproc_X.apply(self.X_train)
            if self.has_Y:
                self.preproc_Y = PreProcess(f_pp,  a_pp, is_Y = True)
                self.Y_train = self.preproc_Y.apply(self.Y_train)
                
        # preparing the testing pre-processing functions
                            
        if f_pp_test is not None:
            self.preproc_X.f_pp = f_pp_test
            self.preproc_X.a_pp = a_pp_test
            if self.has_Y:
                self.preproc_Y.f_pp = f_pp_test
                self.preproc_Y.a_pp = a_pp_test
                
        # aligning the rows of X and Y, if necessary            
        if self.has_Y:
            self.X_train, self.Y_train = align_dfs_by_rows(self.X_train,
                                                            self.Y_train)
        
        # saving the tags
        self.tags_X = self.X_train.columns.to_list()
        if self.has_Y:
            self.tags_Y = self.Y_train.columns.to_list()
          
    ###########################################################################

    def train (self, lim_conf = 0.99, delete_training_data = False):
        """
        Performs the model training. 
        Must be called after the pre_train function.

        Parameters
        ----------
        lim_conf: float, optional
            Confidence limit for the detection index.
        delete_training_data: boolean, optional
            If True, the data is deleted at the end of training.
            Useful to save memory.
        """        
        
        self.lim_conf = lim_conf

        start_train = time.time()
        
        # core of the training algorithm!
        self.train_core()
        
        # predicting! (or reconstructing!)
        train_pred = self.map_from_X (self.X_train.values)

        end_train = time.time()  
        
        self.train_time = end_train - start_train
        
        # storing results and calculating SPE
        
        if self.has_Y:        
            self.Y_train_pred = pd.DataFrame(train_pred,
                                            index=self.Y_train.index,
                                            columns = self.Y_train.columns)  
            
            self.SPE_train = np.sum((self.Y_train.values-
                                    self.Y_train_pred.values)**2,
                                    axis=1)
        else:
            self.X_train_pred = pd.DataFrame(train_pred,
                                            index=self.X_train.index,
                                            columns = self.X_train.columns) 

            self.SPE_train = np.sum((self.X_train.values-
                                    self.X_train_pred.values)**2,
                                    axis=1)
            
        # calculating indices and detection limits

        self.SPE_mean = self.SPE_train.mean()
        
        iSPE = np.sort(self.SPE_train)
        self.limSPE = iSPE[int(lim_conf*self.X_train.shape[0])]
        
        self.limSPE_train = copy.deepcopy(self.limSPE)
        
        # denormalizing
        
        if self.has_Y and self.preproc_Y is not None:
            if 'normalize' in self.preproc_Y.f_pp:
                self.Y_train_pred_orig = \
                    self.preproc_Y.back_to_units(self.Y_train_pred)
            else:
                self.Y_train_pred_orig = self.Y_train_pred
        else:
            if 'normalize' in self.preproc_X.f_pp:
                self.X_train_pred_orig = \
                    self.preproc_X.back_to_units(self.X_train_pred)
            else:
                self.X_train_pred_orig = self.X_train_pred

        # indexes of the training set
        self.index_train = self.X_train.index
        
        self.SPE_train = pd.Series(self.SPE_train, index = self.index_train)
        
        # deleting training data, if applicable
        if delete_training_data:
            del self.X_train, self.X_train_orig, \
                self.SPE_train, self.index_train
            if self.has_Y:
                del self.Y_train, self.Y_train_orig, \
                    self.Y_train_pred, self.Y_train_pred_orig
            else:
                del self.X_train_pred, self.X_train_pred_orig    
    
    ###########################################################################
    
    def hyperparameter_tuning (self, params, n_trials = 20, lim_conf = 0.99,
                               percent_validation = 0.2,
                               n_splits = None, delete_training_data = False):
        """
        Performs hyperparameter tuning using the Optuna library.

        Parameters
        ----------
        params: pandas.DataFrame
            Contains the possibilities to be tested and the types of 
            parameters. Must be defined as: 
            pd.DataFrame({'possibilities': [list_possibilities1, 
                                            list_possibilities2, ...],
                          'types': [str_type1, 
                                    str_type2, ...]},
                         index = [str_param1_name, str_param2_name, ...])
        n_trials: int, optional
            Number of trials in the optimization performed by Optuna.
        lim_conf: float, optional
            Confidence limit for the detection index.
        percent_validation: float (0<value<1), optional
            Percentage of the data to be separated for use 
            in internal validation, if no cross-validation is performed.
        n_splits: int, optional
            Number of sets to be used in cross-validation.
            If not None, the value of percent_validation is disregarded.
        delete_training_data: boolean, optional
            If True, the data is deleted at the end of training.
            Useful to save memory.
        """  
        def objective (trial):

            suggestions = {}
            
            for index, row in params.iterrows():
                if row['type'] == 'uniform':
                    sug = trial.suggest_uniform(index, 
                                                row['possibilities'][0], 
                                                row['possibilities'][1])                  
                if row['type'] == 'loguniform':
                    sug = trial.suggest_loguniform(index, 
                                                row['possibilities'][0], 
                                                row['possibilities'][1])  
                if row['type'] == 'discrete_uniform':
                    sug = trial.suggest_discrete_uniform(index, 
                                                       row['possibilities'][0], 
                                                       row['possibilities'][1],
                                                       row['possibilities'][2])   
                if row['type'] == 'int':
                    sug = trial.suggest_int(index, 
                                            int(row['possibilities'][0]), 
                                            int(row['possibilities'][1]))
                if row['type'] == 'categorical':
                    sug = trial.suggest_categorical(index, 
                                                    row['possibilities']) 
                suggestions[index] = sug
            
            self.set_hyperparameters(suggestions)
            
            if n_splits is not None:
            
                kf = sklearn.model_selection.KFold(n_splits = n_splits)
                            
                SPEs = []
                
                for train_index, test_index in kf.split(self.X_train):
                    autocopy = copy.deepcopy(self)
                    autocopy.pre_train(X_train =
                                       self.X_train.iloc[train_index,:],
                                       Y_train = 
                                       self.Y_train.iloc[train_index,:])
                    autocopy.train(lim_conf, delete_training_data = True)
                    
                    autocopy.pre_test(X_test = self.X_train.iloc[test_index,:],
                                      Y_test = self.Y_train.iloc[test_index,:])
                    
                    autocopy.test()
                    SPEs.append(np.mean(autocopy.SPE_test))

                del autocopy
                return np.mean(SPEs)
            
            else:
                
                n = int(self.X_train.shape[0]*percent_validation)
                
                autocopy = copy.deepcopy(self)
                
                autocopy.pre_train(X_train = self.X_train.iloc[n:,:],
                                    Y_train = self.Y_train.iloc[n:,:])
                autocopy.train(lim_conf, delete_training_data = True)
                autocopy.pre_test(X_test = self.X_train.iloc[:n,:],
                                    Y_test = self.Y_train.iloc[:n,:])
                autocopy.test()
            
                mean_spe = np.mean(autocopy.SPE_test)
                del autocopy
                return mean_spe
        
        self.hyperparemeter_study = optuna.create_study()
        self.hyperparemeter_study.optimize(objective, n_trials = n_trials)
        
        print(self.hyperparemeter_study.best_params)
        
        self.set_hyperparameters(self.hyperparemeter_study.best_params)
                   
    ###########################################################################

    def pre_test (self, X_test, Y_test = None, 
                  count_window_size = 0, count_limit = 1,
                  f_pp = None, a_pp = None):
        '''   
        Receives a window of data and prepares it for the model testing.
        
        Parameters
        ----------
        X_test: pandas.DataFrame, pandas.Series or numpy.ndarray
            Window of data or observation X needed to perform the analysis.
        Y_test: pandas.DataFrame, pandas.Series or numpy.ndarray, optional
            Window of data or observation Y needed to perform the analysis.
        count_window_size: int, optional
            Window size used in the count alarm calculation.
        count_limit: int, optional
            Limit of points to be considered in the window 
            for the count alarm to trigger. 
        f_pp: list, optional
            List containing strings with names of functions to be used 
            in pre-processing (the functions are defined in the 
            PreProcess class).
        a_pp: dict, optional
            Dictionary containing the parameters to be provided
            to each function to perform pre-processing of the data, in
            the format {'functionname__argname': argvalue, ...}
        '''
        
        # window sizes for alarms:
                        
        self.count_limit = count_limit
        self.count_window_size = count_window_size

        # storing complete testing set
                                
        if (isinstance(X_test, np.ndarray) and X_test.shape == ()):
            self.X_test = pd.DataFrame(X_test[None])
        elif len(X_test.shape)==1 and ((isinstance(X_test, np.ndarray))):
                self.X_test = pd.DataFrame(X_test).T
        else:
            self.X_test = pd.DataFrame(X_test)  
            
        if self.has_Y:
            if (isinstance(Y_test, np.ndarray) and Y_test.shape == ()):
                self.Y_test = pd.DataFrame(Y_test[None])
            elif len(Y_test.shape)==1 and ((isinstance(Y_test, np.ndarray))):
                    self.Y_test = pd.DataFrame(Y_test).T
            else:
                self.Y_test = pd.DataFrame(Y_test)

        # data set that will keep the original units       
                
        self.X_test_orig = copy.deepcopy(self.X_test)
        if self.has_Y:
            self.Y_test_orig = copy.deepcopy(self.Y_test) 

        # applying pre-processing functions

        if f_pp is not None:

            if not hasattr(self, 'preproc_X'):
                self.preproc_X = PreProcess(f_pp, a_pp)
            else:
                self.preproc_X.f_pp = f_pp
                self.preproc_X.a_pp = a_pp

            if self.has_Y:

                if not hasattr(self, 'preproc_Y'):
                    self.preproc_Y = PreProcess(f_pp, a_pp, is_Y = True)
                else:
                    self.preproc_Y.f_pp = f_pp
                    self.preproc_Y.a_pp = a_pp

        if hasattr(self, 'preproc_X'):
            self.X_test = self.preproc_X.apply(self.X_test, 
                                                train_or_test = 'test')

        if hasattr(self, 'preproc_Y'):
            self.Y_test = self.preproc_Y.apply(self.Y_test, 
                                                train_or_test = 'test')  

        # aligning the rows of X and Y, if necessary
        if self.has_Y:
            self.X_test, self.Y_test = align_dfs_by_rows(self.X_test,
                                                         self.Y_test)

        # test tags   
            
        self.tags_test_X = [t for t in self.tags_X
                            if t in self.X_test.columns.tolist()]
        
        self.X_test = self.X_test[self.tags_test_X]

        if self.has_Y:

            self.tags_test_Y = [t for t in self.tags_Y 
                                if t in self.Y_test.columns.tolist()]
                
            self.Y_test = self.Y_test[self.tags_test_Y]
                        
    ###########################################################################

    def test (self, redefine_limit = False, delete_testing_data = False):
        """
        Analyzes a window of data, applying a model test.
        Must be called after the pre_test function.
        
        Parameters
        ----------
        redefine_limit: boolean, optional
            Indicator of redefinition or not of the detection limit during
            testing.
        delete_testing_data: boolean, optional
            If True, the data is deleted at the end of testing.
            Useful to save memory.
        """

        start_test = time.time()
        
        # predicting!
        test_pred = self.map_from_X(self.X_test.values)

        end_test = time.time()
        
        self.test_time = end_test - start_test

        # storing results and calculating SPE
    
        if self.has_Y:
            self.Y_test_pred = pd.DataFrame(test_pred,
                                            index=self.Y_test.index,
                                            columns = self.Y_test.columns)
            self.SPE_test = np.sum((self.Y_test.values-
                                    self.Y_test_pred.values)**2,
                                    axis=1)
            
        else:
            self.X_test_pred = pd.DataFrame(test_pred,
                                            index=self.X_test.index,
                                            columns = self.X_test.columns)
            self.SPE_test = np.sum((self.X_test.values-
                                    self.X_test_pred.values)**2,
                                    axis=1)   
        
        # redefining the limit, for the validation case
        
        if redefine_limit:
            iSPE = np.sort(self.SPE_test)
            self.limSPE = iSPE[int(self.lim_conf*self.X_test.shape[0])]
        
        # calculations of the alarms
        
        if self.count_window_size != 0:
            self.alarmCount = np.zeros(len(self.SPE_test))

        self.alarmOutlier = detecOutlier(self.SPE_test, 
                                         self.limSPE)

        # if more window alarms are defined in the package, 
        # i_min must be the minumum size among them
        i_min = self.count_window_size
        
        for i in range(i_min, len(self.SPE_test)):
                        
            if self.count_window_size != 0:
                if i >= self.count_window_size:
                    self.alarmCount[i] = \
                    detecOutlier(self.SPE_test[i-self.count_window_size:(i+1)], 
                                 self.limSPE, count = True, 
                                 count_limit = self.count_limit)
            
        # adjusting the size of the prediction and denormalizing    
        if self.has_Y:
            if 'normalize' in self.preproc_Y.f_pp:
                self.Y_test_pred_orig = \
                    self.preproc_Y.back_to_units(self.Y_test_pred)
            else:
                self.Y_test_pred_orig = self.Y_test_pred
        else:
            if 'normalize' in self.preproc_X.f_pp:
                self.X_test_pred_orig = \
                    self.preproc_X.back_to_units(self.X_test_pred)
            else:
                self.X_test_pred_orig = self.X_test_pred
        
        self.SPE_test = pd.Series(self.SPE_test, 
                                index = self.X_test.index)

        self.alarms = {}

        self.alarmOutlier = pd.Series(self.alarmOutlier,
                                    index = self.X_test.index)
        self.alarms['alarmOutlier'] = self.alarmOutlier
        if self.count_window_size != 0:
            self.alarmCount = pd.Series(self.alarmCount,
                                        index = self.X_test.index)
            str_count = f'alarmCount WS={self.count_window_size}, '\
                    f'limCount = {self.count_limit}'
            self.alarms[str_count] = self.alarmCount

        # deleting testing data, if applicable
        if delete_testing_data:
            del self.X_test, self.X_test_orig, \
                self.SPE_test, self.alarms, self.alarmOutlier
            if self.has_Y:
                del self.Y_test, self.Y_test_orig, \
                    self.Y_test_pred, self.Y_test_pred_orig
            else:
                del self.X_test_pred, self.X_test_pred_orig
            if hasattr(self, 'alarmCount'): del self.alarmCount

    ###########################################################################
        
    def plot_SPE (self, ax = None, train_or_test = 'train', logy = True,
                  legends = True, plot_alarm_outlier = True):
        """
        Plotting the temporal evolution of SPE.

        Parameters
        ----------
        ax: matplotlib.axes._subplots.AxesSubplot, optional
            Axis on which the graph will be plotted.
        train_or_test: string, optional
            Indicates whether to plot the graph for 'train' or 'test'.
        logy: boolean, optional
            Indicates whether the y-axis scale should be logarithmic (True) or 
            linear (False).
        legends: boolean, optional
            If the graph should display legends.
        plot_alarm_outlier: boolean, optional
            If the alarmOutlier should be plotted.
        """
        
        if ax is not None:
            pass
        else:
            fig, ax = plt.subplots()
            
        if train_or_test == 'train':
            SPE = self.SPE_train
            limSPE = self.limSPE_train
        elif train_or_test == 'test':
            SPE = self.SPE_test
            limSPE = self.limSPE
        
        SPE.plot(ax=ax, logy = logy, ls='',
                marker='.', label='SPE')
        
        if train_or_test == 'test':
            for label, alarm in self.alarms.items():
                if not plot_alarm_outlier and label=='alarmOutlier':
                    continue
                (ax.get_ylim()[1]*alarm).replace(0, np.nan).plot(ax = ax, 
                                                                logy = logy, 
                                                                ls = '',
                                                                marker = '.', 
                                                                alpha = 0.5,
                                                                label = label)
                    
        ax.axhline(limSPE, color='red', ls = '--',
                label='%.0f%% Confidence Limit' %(self.lim_conf*100))
        
        if legends:
            ax.legend(fontsize=12);
                        
    ###########################################################################

    def plot_predictions (self, ax = None, train_or_test = 'train', 
                          X_pred_to_plot = None, metrics = None):
        """
        Plotting the temporal evolution of the predictions along with the
        respective true values.

        Parameters
        ----------
        ax: matplotlib.axes._subplots.AxesSubplot, optional
            Axis on which the graph will be plotted.
        train_or_test: string, optional
            Indicates whether to plot the graph for 'train' or 'test'.
        X_pred_to_plot: string, optional
            In case the model is a reconstruction model 
            (i.e., self.has_Y = False), indicates which column of X to plot 
            along with the prediction.
        metrics: list of functions, optional
            Functions for calculating metrics to be displayed in the title of 
            the graph.
        """ 

        if ax is not None:
            pass
        else:
            fig, ax = plt.subplots()
        
        if self.has_Y:
        
            if train_or_test=='train':
                true = self.Y_train_orig
                pred = self.Y_train_pred_orig
            elif train_or_test=='test':
                true = self.Y_test_orig
                pred = self.Y_test_pred_orig
            
            regression_or_reconstruction = 'Regression'
            
        else:
            
            if X_pred_to_plot is not None:

                if train_or_test=='train':
                    true = self.X_train_orig[X_pred_to_plot]
                    pred = self.X_train_pred_orig[X_pred_to_plot]
                elif train_or_test=='test':
                    true = self.X_test_orig[X_pred_to_plot]
                    pred = self.X_test_pred_orig[X_pred_to_plot]
                
                regression_or_reconstruction = 'Reconstruction'
                
            else:

                raise RuntimeError('The model has no Y. \
                                Specify which variable in X should be plotted')
                
        true.plot(ax=ax)
        pred.plot(ax=ax,ls='--')
        
        ax.legend(['Measurement',regression_or_reconstruction])
        
        tags = pd.DataFrame(true).columns.to_list()
                
        title = str(tags)
        
        if metrics is not None:
            
            if not isinstance(metrics, list): 
                metrics = [metrics]
                                
            true, pred = align_dfs_by_rows(true.dropna(), pred)
            
            for metric in metrics:
                title = (title+
                        f'\n{metric.__name__}: {metric(true, pred):.2f}')
        
        ax.set_title(title)
    
    ###########################################################################

    def fit (self, X_train, Y_train = None,
             f_pp = ['remove_empty_variables',
                     'ffill_nan',
                     'remove_frozen_variables',
                     'normalize'],
             a_pp = None,
             f_pp_test = ['replace_nan_with_values',
                          'normalize'],
             a_pp_test = None,
             lim_conf = 0.99,
             delete_training_data = False,
             redefine_limit = False,
             frac_val = 0.15,
             tune = False,
             params = None,
             params_types = None,
             params_possibilities = None,
             n_trials = 20):
        """
        Performs the complete pipeline of model training,
        sequentially executing the 'pre_train' and 'train' methods.

        Parameters
        ----------
        X_train: pandas.DataFrame or numpy.ndarray
            Window of X data used in training.
        Y_train: pandas.DataFrame or numpy.ndarray, optional
            Window of Y data used in training.
        f_pp: list, optional
            List containing strings with names of functions to be used 
            in pre-processing the training data (the functions are defined
            in the PreProcess class, in the BibMon_Tools.py file).
        f_pp_test: list, optional
            List containing strings with names of functions to be used 
            in pre-processing the testing data (the functions are defined 
            in the PreProcess class, in the BibMon_Tools.py file).
        a_pp: dict, optional
            Dictionary containing the parameters to be provided
            to each function to perform pre-processing of the training data, in
            the format {'functionname__argname': argvalue, ...}.
        a_pp_test: dict, optional
            Dictionary containing the parameters to be provided
            to each function to perform pre-processing of the testing data, in
            the format {'functionname__argname': argvalue, ...}.
        lim_conf: float, optional
            Confidence limit for the detection index.
        delete_training_data: boolean, optional
            If True, the data is deleted at the end of training.
            Useful to save memory.
        redefine_limit: boolean, optional
            Indicator of redefinition or not of the detection limit using
            a validation period taken from the training data itself.
        frac_val: float, optional
            Fraction of the data used for validation. 
            0<frac_val<1.
            Only used if redefine_limit==True.
        tune: boolean, optional
            Indicator of automatic hyperparameter tuning.
        params: string or list of strings, optional
            Name(s) of the parameter(s) to be tuned.
        params_types:  string or list of strings
            Type(s) of the parameter(s) to be tuned.
        params_possibilities: list, optional
            Possibilities to be tested for each parameter.
            It must be a list containing the possibilities 
            (in case of only one parameter), or a list containing the lists
            for each possibility.
            Possibilities must be provided according to the type of
            the parameter, as specified in the Optuna library API.
        n_trials: int, optional
            Number of iterations in the hyperparameter search optimization.
        """

        X_train = pd.DataFrame(X_train)
        Y_train = pd.DataFrame(Y_train)

        if redefine_limit:
            
            n_trn =  int(X_train.shape[0]*(1-frac_val))
            X_val = X_train.iloc[n_trn:]
            X_train = X_train.iloc[:n_trn]
            
            if Y_train is not None:

                Y_val = Y_train.iloc[n_trn:]
                Y_train = Y_train.iloc[:n_trn]
            
            else:
                
                Y_val = None
            
        if tune:
            self.pre_train(X_train, Y_train, 
                f_pp = f_pp,
                a_pp = a_pp,
                f_pp_test = f_pp_test,
                a_pp_test  = a_pp_test)
            if not isinstance(params, list):
                params = [params]
            if not isinstance(params_types, list):
                params_types = [params_types]
            if not isinstance(params_possibilities, list):
                print('params_possibilities must be a list!')
            else:
                if not isinstance(params_possibilities[0], list):
                    params_possibilities = [params_possibilities]
            params_df = pd.DataFrame({'possibilities': params_possibilities,
                                      'type': params_types},
                                     index = params) 
            self.tuning(n_trials = n_trials, params = params_df, 
                        lim_conf = lim_conf, 
                        delete_training_data = delete_training_data)
            
        self.pre_train(X_train, Y_train, 
                        f_pp = f_pp,
                        a_pp = a_pp,
                        f_pp_test = f_pp_test,
                        a_pp_test  = a_pp_test)

        self.train(lim_conf = lim_conf, 
                   delete_training_data = delete_training_data)

        if redefine_limit:
            self.pre_test(X_val, Y_val)
            self.test(redefine_limit = True)
        if f_pp_test:   
            # save value for use in predict if necessary
            self.f_pp_test = f_pp_test 
        if a_pp_test:  
             # save value for use in predict if necessary
            self.a_pp_test = a_pp_test

    ###########################################################################
        
    def predict (self, 
                 X_test, 
                 Y_test = None,  
                 count_window_size = 0, 
                 count_limit = 1, 
                 f_pp = 'fit', 
                 a_pp = 'fit', 
                 delete_testing_data = False,
                 redefine_limit = False):
        '''   
        Performs the complete pipeline of model prediction (testing), 
        sequentially executing the 'pre_test' and 'test' methods.
        
        Parameters
        ----------
        X_test: pandas.DataFrame, pandas.Series or numpy.ndarray
            Window of data or observation X needed to perform the analysis.
        Y_test: pandas.DataFrame, pandas.Series or numpy.ndarray, optional
            Window of data or observation Y needed to perform the analysis.
        count_window_size: int, optional
            Window size used in count alarm calculation. 
        count_limit: int, optional
            Limit of points to be considered in the window 
            for the count alarm to trigger.            
        f_pp: list, optional
            List containing strings with names of functions to be used 
            in pre-processing (the functions are defined in the 
            PreProcess class).
        a_pp: dict, optional
            Dictionary containing the parameters to be provided
            to each function to perform pre-processing of the data, in
            the format {'functionname__argname': argvalue, ...}
        delete_testing_data: boolean, optional
            If True, the data is deleted at the end of testing.
            Useful to save memory.   
        redefine_limit: boolean, optional
            Indicator of redefinition or not of the detection limit during
            testing.
        '''
        
        if f_pp == 'fit':
            if hasattr(self,'f_pp_test'):
                f_pp = self.f_pp_test
            else:
                f_pp = None
        if a_pp == 'fit':
            if hasattr(self,'a_pp_test'):
                a_pp = self.a_pp_test
            else:
                a_pp = None
            
        self.pre_test(X_test, Y_test, count_limit = count_limit, 
                      count_window_size = count_window_size, 
                      f_pp = f_pp, a_pp = a_pp)
        self.test(redefine_limit = redefine_limit,
                  delete_testing_data = delete_testing_data)
        


import numpy as np

###############################################################################

def detecOutlier(data, lim, count = False, count_limit = 1):
    """
    Detects outliers in the given data using a specified limit.

    Parameters
    ----------
    data: array-like
        The input data to be analyzed.
    lim: float
        The limit value used to detect outliers.
    count: bool, optional
        If True, counts the number of outliers 
        exceeding the limit. Default is False.
    count_limit: int, optional
        The maximum number of outliers allowed. 
        Default is 1.

    Returns
    ----------
    alarm: ndarray or int
        If count is False, returns an array indicating 
        the outliers (0 for values below or equal to lim,
        1 for values above lim).
        If count is True, returns the number of outliers 
        exceeding the limit.
    """

    if np.isnan(data).any():
        data = np.nan_to_num(data)

    if count == False:
        alarm = np.copy(data)
        alarm = np.where(alarm<=lim, 0, alarm)
        alarm = np.where(alarm>lim, 1, alarm)
        return alarm
    else:
        alarm = 0
        local_count = np.count_nonzero(data > lim)
        if local_count > count_limit:
            alarm = +1
        return alarm
        



from .generic_model import GenericModel

class NewModel (GenericModel):
    
    ###########################################################################
            
    def __init__ (self):
        """
        Constructor.

        Here, the variable has_Y should be specified.
        This variable indicates whether the model has a 
        separate set of prediction variables Y or not.

        It is recommended to inform model initialization parameters here.
        """
        self.has_Y = False # or True!
        
    ###########################################################################
    
    def load_model (self,
                         limSPE, SPE_mean, lagX, count_window_size,
                         Mux, SDx, Muy = None, SDy = None,
                         # potential additional parameters,
                         ):
        
        super().load_model (limSPE, SPE_mean, count_window_size, 
                            Mux, SDx, Muy, SDy)
        
        """
        Receives parameters from a previously trained model for
        making predictions and tests without the need for training.
        
        !!
        Below, you should store additional parameters as attributes 
        of the class; if there are no additional parameters, 
        this method can be deleted.
        !!
    
        Parameters:
        ----------
        Mux: pandas.Series
            Means of the X variables in the training period.
        SDx: pandas.Series
            Standard deviations of the X variables in
            the training period.
        Muy: pandas.Series, optional
            Means of the Y variables in the training period.
        SDy: pandas.Series, optional
            Standard deviations of the Y variables in the training period.                         
        limSPE: float
            Detection limit and mean of the SPE.
        SPE_mean: float
            Mean of the SPE.
        count_window_size: int
            Window sizes used in count alarms calculation.
        """

    ###########################################################################
        
    def train_core (self):
        """
        The core of the training algorithm, that is,
        all the necessary steps between pre_train() and
        the calculation of the prediction in training.
        """

    ###########################################################################

    def map_from_X (self, X): 
        """
        Receives a data matrix X and returns a matrix of predicted 
        or reconstructed values.

        Parameters
        ----------
        X: numpy.array
            Window X of data for prediction or reconstruction.  
            
        Returns
        ----------                
        : numpy.array
            Reconstructed X (or predicted Y).
        """  


import sys
import numpy as np
import scipy as sp
from ._generic_model import GenericModel

###############################################################################

def p_norm(x, p=2.0):
    """
    Parameters
    ----------
    x: array
        Vector to have the norm calculated;
        norm_p(x) = ( sum (|x_i|^p) ) ^(1/p)
    """
    return np.sum((np.abs(x)**p))**(1/p)

###############################################################################

def calculate_vector_similarity(x, y, p=2.0,functional_form='rbf',
                                gamma=1.0, eta=1e-10):
    """
    Parameters
    ----------        
    x, y: array
        Vectors to have the similarity calculated.
    p: float, optional
        p_value for evaluating the function p_norm(x,p_value)
    functional_form: string, optional
        rbf = radial basis function; 
        ies = inverse euclidean similarity; 
        iqk = inverse quadratic kernel; 
        exp_kernel = exponential kernel; 
        cauchy_kernel = cauchy kernel.
    gamma: float, optional
        Parameter used in the functional form.
    eta: float, optional
        Minimum similarity value to be returned.
        
    Returns
    ----------        
    : float
        Similarity.
    """
    dif = x-y
    
    if functional_form=='rbf':
        sim = np.exp(-gamma*p_norm(dif,p)**2)
               
    elif functional_form=='ies':
        sim = 1./(1+gamma*p_norm(dif,p))
        
    elif functional_form=='iqk':
        sim = 1./(np.sqrt(1+(gamma*p_norm(dif,p))**2))

    elif functional_form=='exp_kernel':
        sim = np.exp(-gamma*p_norm(dif,p))
        
    elif functional_form=='cauchy_kernel':
        sim = 1./(1+(gamma*p_norm(dif,p))**2)

    if sim == 0.: sim = eta

    return sim

###############################################################################

def calculate_matrix_similarity(A, B, p=2.0,functional_form='rbf',
                                gamma=1.0, eta=1e-10):
    
    """
    Parameters
    ----------        
    A: numpy.array
        Left matrix (vectors in row).
    B: numpy.array
        Right matrix (vectors in column).
    p: float, optional
        p_value for evaluating the function p_norm(x,p_value)
    functional_form: string, optional
        rbf = radial basis function; 
        ies = inverse euclidean similarity; 
        iqk = inverse quadratic kernel; 
        exp_kernel = exponential kernel; 
        cauchy_kernel = cauchy kernel.
    gamma: float, optional
        Parameter used in the functional form.
    eta: float, optional
        Minimum similarity value to be returned.
        
    Returns
    ----------        
    : numpy.array
        Similarity matrix.
    """
    if A.ndim < 2:
        A = np.reshape(A, (len(A), -1))
    if B.ndim < 2:
        B = np.reshape(B, (len(B), -1))
    
    rows = np.shape(A)[0]
    columns = np.shape(B)[1]
    C = np.zeros((rows,columns))
        
    for i in range(rows):
        for j in range(columns):
            C[i,j]=calculate_vector_similarity(A[i,:],B[:,j], 
                                               p,
                                               functional_form,gamma,eta)
    
    return C

###############################################################################

class SBM (GenericModel):
    """
    Similarity-based method (SBM).
    
    For details on the technique, see the papers:
        
        * Marins et al. (2018) - Improved similarity-based modeling for the 
          classification of rotating-machine failures, 
          http://www02.smt.ufrj.br/~sergioln/papers/IJ28.pdf
        * Ribeiro (2018) - Similarity-based methods for machine diagnosis, 
        http://www.pee.ufrj.br/index.php/pt/producao-academica/teses-de-doutorado/tese-1/2016033299-similarity-based-methods-for-machine-diagnosis/file
    
    Parameters
    ----------  
        p: float, optional
            p-value for the definition of the norm.
        functional_form: string, optional
            Functional form to be used in the similarity calculation.
            rbf = radial basis function; 
            ies = inverse euclidean similarity; 
            iqk = inverse quadratic kernel; 
            exp_kernel = exponential kernel; 
            cauchy_kernel = cauchy kernel.
        gamma: float, optional
            Parameter present in the various functional forms 
            of similarity.
        eta: float, optional
            Minimum value to be returned in similarity calculations.
        train_method: string, optional
            Training method. Options: 'all_archetypes' and
            'geometrical_median'.
        tau: float, optional
            Similarity threshold.
        verbose: boolean, optional
            Whether to print information during execution.
    """    

    ###########################################################################

    def __init__ (self, p = 2.0, functional_form = 'rbf', gamma = 1.0, 
                  eta = 1e-10, train_method = 'geometrical_median',
                  tau = 1.e-10, verbose = False):

        self.has_Y = False

        self.p = p
        self.function = functional_form
        self.gamma = gamma
        self.eta=eta
        self.train_method = train_method
        self.tau = tau
        self.verbose = verbose
        
        self.name = 'SBM'
        
    ###########################################################################
    
    def train_core (self):
     
        M = self.X_train.values
        if self.verbose:
            print('#########################################')
            print('Training SMB')
            print('')
            print('Full memory matrix with ',np.shape(M)[0],
                  ' archetypal states')
            print('')
            print('applying the ' , self.train_method, 
                  'method of training...' )
            print('')
        
        if self.train_method=='all_archetypes':
            self.D = M
        #elif method == 'original':
        
        elif self.train_method == 'geometrical_median':
            
            mean = np.mean(M,axis=0)
            
            minor = sys.maxsize
            for i in range(np.shape(M)[0]):
                dist_i = p_norm((M[i,:]-mean), p=2.0)
                if dist_i<minor:
                    minor = dist_i
                    median = M[i,:]
            
            self.D = median
            for i in range(np.shape(M)[0]):
                x = M[i,:].T
                C = calculate_matrix_similarity(self.D, x,self.p,
                                                self.function,
                                                self.gamma,self.eta)
                sim_mean = np.mean(C)
                if sim_mean<self.tau:
                    self.D=np.vstack((self.D, x.T))
                    
        self.Dt = self.D.T
        self.invsimDDt = sp.linalg.inv(calculate_matrix_similarity(self.D, 
                                                                   self.Dt,
                                                                   self.p,
                                                                 self.function,
                                                                   self.gamma,
                                                                   self.eta))

        if self.verbose:

            print('SBM trained - D matrix with ', 
                  np.shape(self.D)[0],' archetypal states')
            print('')
            print('#########################################')
        
        
    #########################
   
    def map_from_X(self, X):
        
        Xt = X.T
        n = np.shape(X)[0] # quantity of current states
        
        X_est = np.zeros(np.shape(X))
        
        for i in range(n):
            w = self.invsimDDt@calculate_matrix_similarity(self.D, Xt[:,i],
                                           self.p,self.function,
                                           self.gamma,self.eta)
            normalizer = np.sum(w)
            w = w/normalizer
            
            x = self.Dt@w 
            X_est[i,:]= x.T
            
        return X_est
        



import numpy as np
import matplotlib.pyplot as plt

from ._generic_model import GenericModel

###############################################################################

class PCA (GenericModel):
    """
    Principal Component Analysis.
    
    For details on the technique, see https://doi.org/10.3390/pr12020251
    
    Parameters
    ----------
    ncomp: int or float
           float: number between 0.0 and 1.0 that corresponds to the minimum 
                  fraction of accumulated variance for component selection;
           int: defines the number of components.
    """

    ###########################################################################
            
    def __init__ (self, ncomp=0.9):

        self.has_Y = False
        self.ncomp = ncomp
        self.name = 'PCA'
        
    ###########################################################################
    
    def load_model (self,limSPE, SPE_mean, count_window_size,
                    Mux, SDx, S, V, n):
        
        """
        Receives parameters from a previously trained model to
        perform predictions and tests without the need for training.
    
        Parameters
        ----------
        Mux: pandas.Series
            Means of the X variables in the training period.
        SDx: pandas.Series
            Standard deviations of the X variables in the training period.
        Mux: pandas.Series, optional
            Means of the Y variables in the training period.
        SDx: pandas.Series, optional
            Standard deviations of the Y variables in the training period.                         
        limSPE: float
            Detection limit and mean of the SPE.
        SPE_mean: float
            Mean of the SPE.
        count_window_size: int
            Window sizes used in the count alarms calculation.        
        S: numpy.array
            Specific parameter of PCA.
        V: numpy.array
            Specific parameter of PCA.
        n: int
            Specific parameter of PCA.        
        """            
                
        super().load_model (limSPE, SPE_mean,
                            count_window_size, Mux, SDx)
        
        self.S = np.array(S)
        self.V = np.array(V)
        self.n = n
        
        self.pv = S/np.sum(S)
        self.pva = np.cumsum(S)/sum(S)

    ###########################################################################
        
    def train_core (self):

        self.Sxx_train = np.cov(self.X_train, rowvar=False)
        _, self.S, self.Vh = np.linalg.svd(self.Sxx_train)
        self.V = self.Vh.T
        self.pv = self.S/np.sum(self.S)
        self.pva = np.cumsum(self.S)/sum(self.S)

        if self.ncomp> 0 and self.ncomp < 1:
            self.n = np.where(self.pva>self.ncomp)[0][0]+1
        elif self.ncomp >= 1:
            self.n = self.ncomp
        else:
            # Error: unkown self.n
            self.n = None
                
    ###########################################################################

    def map_from_X (self, X): 

        X_train_proj = X@self.V[:,:self.n]
        return X_train_proj@self.V[:,:self.n].T
    
    ###########################################################################
        
    def plot_cumulative_variance (self, ax = None):
        """
        Plots the cumulative variance.
    
        Parameters
        ----------
        ax: matplotlib.axes._subplots.AxesSubplot, optional
            Axis on which the graph will be plotted.
        """                
        if ax is not None:
            pass
        else:
            fig, ax = plt.subplots()
        
        ax.bar(np.arange(len(self.pv)), self.pv)
        ax.plot(np.arange(len(self.pv)), self.pva)
        ax.set_xlabel('Number of components')
        ax.set_ylabel('Data variance')


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.inspection import permutation_importance

from ._generic_model import GenericModel

###############################################################################

class sklearnRegressor (GenericModel):
    """
    Interface for sklearn regressors.
            
    Parameters
    ----------
    regressor: any regressor that uses the sklearn interface. 
        For example:
            * sklearn.svm.classes.SVR,
            * sklearn.ensemble.forest.RandomForestRegressor,
            * sklearn.neural_network.multilayer_perceptron.MLPRegressor,
            * etc....
    permutation_importance: boolean, optional
        Whether permutation variable importance should be calculated.    
        """     

    ###########################################################################

    def __init__ (self, regressor, permutation_importance = False):

        self.has_Y = True       
        self.regressor = regressor
	
        self.name = self.regressor.__class__.__name__
        
        self.permutation_importance = permutation_importance
    
    ###########################################################################
        
    def train_core (self):
        
        self.regressor.fit(self.X_train.values,
                           self.Y_train.values.squeeze())
        
        if self.permutation_importance:
            
            res = permutation_importance(self.regressor, 
                                         self.X_train.values, 
                                         self.Y_train.values.squeeze(),
                                         n_repeats=10)
            
            self.regressor.perm_feature_importances_ = res.importances_mean

    ###########################################################################

    def map_from_X (self, X):
        
        return self.regressor.predict(X)
    
    ###########################################################################
    
    def set_hyperparameters (self, params_dict):   
        
        for key, value in params_dict.items():
            setattr(self.regressor, key, value)
            
    ###########################################################################
            
    def update_importances(self):
        """
        Calculates permutation importances of the variables.
        """          

        res = permutation_importance(self.regressor, 
                                     self.X_test.values, 
                                     self.Y_test.values.squeeze(),
                                     n_repeats = 10)
            
        self.regressor.perm_feature_importances_ = res.importances_mean        
            
    ###########################################################################

    def plot_importances(self, n = None, permutation_importance = False):
        
        """
        Plots the permutation importances of the variables.
    
        Parameters
        ----------
        n: int, optional
            Maximum number of variables to be plotted.
        permutation_importance: boolean, optional
            If permutation importances should be prioritized over coefficients
            in linear models.
        """          
        
        model = self.regressor
        
        if hasattr(model,'coef_'):
            imp = model.coef_
        elif hasattr(model,'feature_importances_'):
            imp = model.feature_importances_
        elif (hasattr(model,'perm_feature_importances_')):
            imp = model.perm_feature_importances_

        if ((hasattr(model,'coef_') or
            hasattr(model,'feature_importances_')) and permutation_importance): 
            if hasattr(model, 'perm_feature_importances_'):
                imp = model.perm_feature_importances_
          
        if not (hasattr(model,'coef_') or
                hasattr(model,'feature_importances_') or
                hasattr(model, 'perm_feature_importances_')):
            print('There are no importances calculated for this model.')
            return
        
        tags = self.X_train.columns
        
        if n is not None:
            pass
        else:
            n = len(self.X_train.columns)
    
        fig, ax = plt.subplots(1,2, figsize = (20,4))
    
        coefs = []
        abs_coefs = []
    
        coefs = (pd.Series(imp, index = tags))
        coefs.plot(use_index=False, ax=ax[0]);
        abs_coefs = (abs(coefs)/(abs(coefs).sum()))
        abs_coefs.sort_values(ascending=False).plot(use_index=False, ax=ax[1],
                                                    marker='.')
    
        ax[0].set_title('Relative variable importances')
        ax[1].set_title('Relative variable importances - \
                        descending order')
    
        abs_coefs_df = pd.DataFrame(np.array(abs_coefs).T,
                                    columns = ['Importances'],
                                    index = tags)
    
        df = abs_coefs_df['Importances'].sort_values(ascending=False)
        
        plt.figure()
        df.iloc[0:n].plot(kind='barh', figsize=(15,0.25*n), legend=False)
        
        return df


import numpy as np
from ._generic_model import GenericModel

###############################################################################

def correct_dimensions(s, targetlength):
    """
    Parameters
    ----------
    s: None, scalar or 1D array
    
    targetlength: int
        Expected length of s.
 
    Returns
    ----------                
    : None if s is None, else numpy vector of length targetlength
    """ 
    if s is not None:
        s = np.array(s)
        if s.ndim == 0:
            # If there is a single value, this command creates
            # a vector of the required size filled with that value.
            s = np.array([s] * targetlength)                                       
        elif s.ndim == 1:
            if not len(s) == targetlength:
                raise ValueError("arg must have length " + str(targetlength))
        else:
            raise ValueError("Invalid argument")
    return s

###############################################################################

def linear(x):
    return x    

###############################################################################

class ESN (GenericModel):
    """
    
    Echo State Networks.
    
    For details on the technique, see the paper 
    by Lemos et al. (2021) - Echo State Network Based Soft Sensor 
    for Monitoring and Fault Detection of Industrial Processes, 
    https://doi.org/10.1016/j.compchemeng.2021.107512

    This code has been modified and adapted from the following repository:
    https://github.com/cknd/pyESN"

    Parameters
    ----------
    n_reservoir: int, optional
        Number of neurons in the reservoir.
    spectral_radius: float, optional
        Spectral radius of the recurrent weight matrix.
    sparsity: float, optional
        Proportion of recurrent weights set to zero.
    noise: float, optional
        Noise added to each neuron (regularization).
    input_shift: float or numpy.array
        Scalar or vector of length n_inputs to be added to each
        input dimension before feeding it to the network.
    input_scaling: float or numpy.array
        Scalar or vector of length n_inputs to be multiplied
        with each input dimension before feeding it to the network.
    teacher_forcing: boolean, optional
        If True, results in an ESN with output layer recursion to the 
        dynamic reservoir.
    teacher_scaling: float, optional 
        Factor applied to the target signal.
    teacher_shift: float, optional
        Additive term applied to the target signal.
    out_activation: func, optional
        Output activation function (applied to the readout).
    inverse_out_activation: func, optional
        Inverse of the output activation function.
    random_state: int or np.rand.RandomState, optional
        Positive integer seed, np.rand.RandomState object,
        or None to use numpy's builting RandomState.
    silent: boolean, optional
        Suppress messages.
    """    
    
    ###########################################################################
    
    def __init__(self, n_reservoir=400,
                 spectral_radius=0.95, sparsity=0.95, noise=0.01, 
                 input_shift=None,
                 input_scaling=None, teacher_forcing=False, 
                 feedback_scaling=None,
                 teacher_scaling=None, teacher_shift=None,
                 out_activation=linear, inverse_out_activation=linear,
                 random_state=None, silent=True):

        
        self.has_Y = True
        
        self.n_reservoir = n_reservoir
        self.spectral_radius = spectral_radius
        self.sparsity = sparsity
        self.noise = noise
        self.input_shift = input_shift
        self.input_scaling = input_scaling

        self.teacher_scaling = teacher_scaling
        self.teacher_shift = teacher_shift

        self.out_activation = out_activation
        self.inverse_out_activation = inverse_out_activation
        self.random_state = random_state

        if isinstance(random_state, np.random.RandomState):
            self.random_state_ = random_state
        elif random_state:
            try:
                self.random_state_ = np.random.RandomState(random_state)
            except TypeError as e:
                raise Exception("Invalid seed: " + str(e))
        else:
            self.random_state_ = np.random.mtrand._rand

        self.teacher_forcing = teacher_forcing
        self.silent = silent
        #self.initweights()
        
        self.name = 'ESN'
        
    ###########################################################################
        
    def initweights(self):       
        """
        Initializes the weights of the network.
        """    
        # initialize recurrent weights:
        # begin with a random matrix centered around zero:
        W = self.random_state_.rand(self.n_reservoir, self.n_reservoir) - 0.5
        # delete the fraction of connections given by (self.sparsity):
        W[self.random_state_.rand(*W.shape) < self.sparsity] = 0
        # compute the spectral radius of these weights:
        radius = np.max(np.abs(np.linalg.eigvals(W)))
        # rescale them to reach the requested spectral radius:
        self.W = W * (self.spectral_radius / radius)

        # random input weights:
        self.W_in = self.random_state_.rand(
            self.n_reservoir, self.n_inputs) * 2 - 1
        # random feedback (teacher forcing) weights:
        self.W_feedb = self.random_state_.rand(
            self.n_reservoir, self.n_outputs) * 2 - 1
  
    ###########################################################################
        
    def _update(self, state, input_pattern, output_pattern):
        """
        Executes one update step,
        i.e., calculates the next state of the network by applying the
        recurrent weights to the last state and feeding in the current input
        and output patterns.

        Parameters
        ----------    
        state: numpy.array
            Last state.
        input_pattern: numpy.array
            Input pattern.        
        state: numpy.array
            Last state.              
        """
        if self.teacher_forcing:
            preactivation = (np.dot(self.W, state)
                             + np.dot(self.W_in, input_pattern)
                             + np.dot(self.W_feedb, output_pattern))
        else:
            preactivation = (np.dot(self.W, state)
                             + np.dot(self.W_in, input_pattern))
        return (np.tanh(preactivation)
              + self.noise * (self.random_state_.rand(self.n_reservoir) - 0.5))
  
    ###########################################################################
        
    def _scale_inputs(self, inputs):
        """
        For each input dimension j: multiply by the j'th entry in
        the input_scaling argument and, subsequently, add the j'th entry
        of the input_shift argument.
        
        Parameters
        ----------    
        inputs: numpy.array
            Inputs.          
        Returns
        ----------    
        inputs: numpy.array
            Inputs in the new scale.        
        """
        if self.input_scaling is not None:
            inputs = np.dot(inputs, np.diag(self.input_scaling))
        if self.input_shift is not None:
            inputs = inputs + self.input_shift
        return inputs
  
    ###########################################################################
        
    def _scale_teacher(self, teacher):
        """
        Multiply the teacher/target signal by the teacher_scaling argument,
        then add the teacher_shift argument.
        
        Parameters
        ----------    
        teacher: numpy.array
            Teacher signal.          
        Returns
        ----------    
        teacher: numpy.array
            Teacher signal in the new scale.          
        """
        if self.teacher_scaling is not None:
            teacher = teacher * self.teacher_scaling
        if self.teacher_shift is not None:
            teacher = teacher + self.teacher_shift
        return teacher
  
    ###########################################################################
        
    def _unscale_teacher(self, teacher_scaled):
        """Inverse operation of the _scale_teacher method.
        
        Parameters
        ----------    
        teacher_scaled: numpy.array
            Teacher signal in the new scale.          
        Returns
        ----------    
        teacher_scaled: numpy.array
            Original teacher signal.          
        """
        
        if self.teacher_shift is not None:
            teacher_scaled = teacher_scaled - self.teacher_shift
        if self.teacher_scaling is not None:
            teacher_scaled = teacher_scaled / self.teacher_scaling
        return teacher_scaled
  
    ###########################################################################
    
    def pre_train (self, *args, **kwargs):
        
        super().pre_train (*args, **kwargs)
        
        self.n_inputs = np.shape(self.X_train)[1]
        self.n_outputs = np.shape(self.Y_train)[1]
        self.input_shift = correct_dimensions(self.input_shift, 
                                              self.n_inputs)
        self.input_scaling = correct_dimensions(self.input_scaling, 
                                                self.n_inputs)
        self.initweights()
    
    ###########################################################################

    def train_core (self):
        """
        Harvest the network's reaction to training data,
        train the readout weights.
        The result is the output of the network on the training data,
        using the trained weights.
        """
        # transform any vectors of shape (x,) into vectors of shape (x,1):
        if self.X_train.ndim < 2:
            self.X_train = np.reshape(self.X_train, (len(self.X_train), -1))
        if self.Y_train.ndim < 2:
            self.Y_train = np.reshape(self.Y_train, (len(self.Y_train), -1))
        # transform input and teacher signal:
        inputs_scaled = self._scale_inputs(self.X_train.values)
        teachers_scaled = self._scale_teacher(self.Y_train.values)

        if not self.silent:
            print("harvesting states...")
            print("")
        # step the reservoir through the given input,output pairs:
        states = np.zeros((self.X_train.shape[0], self.n_reservoir))
        for n in range(1, self.X_train.shape[0]):
            states[n, :] = self._update(states[n - 1], inputs_scaled[n, :],
                                        teachers_scaled[n - 1, :])

        # learn the weights, i.e. find the linear combination of collected
        # network states that is closest to the target output
        if not self.silent:
            print("fitting...")
            print("")
        # we'll disregard the first few states:
        transient = min(int(self.X_train.shape[1] / 10), 100)
        # include the raw inputs:
        extended_states = np.hstack((states, inputs_scaled))
        self.W_out = np.dot(np.linalg.pinv(extended_states[transient:, :]),
                self.inverse_out_activation(teachers_scaled[transient:, :])).T
        
        # remember the last state for later:
        self.laststate = states[-1, :]
        self.lastinput = self.X_train.values[-1, :]
        self.lastoutput = teachers_scaled[-1, :]

        # apply learned weights to the collected states:
        #pred_train = self._unscale_teacher(self.out_activation(
        #    np.dot(extended_states, self.W_out.T)))
        
    ###########################################################################
        
    def map_from_X(self, X, continuation=True):
        """
        Apply the learned weights to the network's reactions to new inputs.

        Parameters
        ----------  
        inputs: numpy.array 
            Inputs of shape (N_test_samples x n_inputs)
        continuation: boolean, optional
            If True, start the network from the last training state.
        Returns
        ----------  
        : numpy.array
            Output activation matrix.
        """
        if X.ndim < 2:
            X = np.reshape(X, (len(X), -1))
        n_samples = X.shape[0]

        if continuation:
            laststate = self.laststate
            lastinput = self.lastinput
            lastoutput = self.lastoutput
        else:
            laststate = np.zeros(self.n_reservoir)
            lastinput = np.zeros(self.n_inputs)
            lastoutput = np.zeros(self.n_outputs)

        X = np.vstack([lastinput, self._scale_inputs(X)])
        states = np.vstack(
            [laststate, np.zeros((n_samples, self.n_reservoir))])
        outputs = np.vstack(
            [lastoutput, np.zeros((n_samples, self.n_outputs))])

        for n in range(n_samples):
            states[
                n + 1, :] = self._update(states[n, :], 
                                         X[n + 1, :], outputs[n, :])
            outputs[n + 1, :] = self.out_activation(np.dot(self.W_out,
                              np.concatenate([states[n + 1, :], X[n + 1, :]])))
       
        return self._unscale_teacher(self.out_activation(outputs[1:]))



import numpy as np
from sklearn.neural_network import MLPRegressor

from ._generic_model import GenericModel

###############################################################################

class Autoencoder (GenericModel):
    
    """
    Autoencoder using sklearn's MLPRegressor interface. 
    For details on the parameters for input, see https://scikit-learn.org/stable/modules/generated/sklearn.neural_network.MLPRegressor.html
    """

    ###########################################################################

    def __init__ (self,hidden_layer_sizes=(2,),activation='relu',solver='adam', 
                  alpha=0.0001, batch_size='auto', learning_rate='constant', 
                  learning_rate_init=0.001, power_t=0.5, max_iter=200, 
                  shuffle=True, random_state=None, tol=0.0001, verbose=False, 
                  warm_start=False, momentum=0.9, nesterovs_momentum=True, 
                  early_stopping=False, validation_fraction=0.1, beta_1=0.9, 
                  beta_2=0.999, epsilon=1e-08, n_iter_no_change=10, 
                  max_fun=15000):

        self.has_Y = False       

        self.regressor = MLPRegressor(hidden_layer_sizes,activation, 
                                      solver=solver, alpha=alpha,
                                      batch_size=batch_size, 
                                      learning_rate=learning_rate,
                                      learning_rate_init=learning_rate_init, 
                                      power_t=power_t,
                                      max_iter=max_iter, shuffle=shuffle,
                                      random_state=random_state, tol=tol, 
                                      verbose = verbose,
                                      warm_start=warm_start, momentum=momentum,
                                      nesterovs_momentum=nesterovs_momentum, 
                                      early_stopping= early_stopping, 
                                      validation_fraction=validation_fraction, 
                                      beta_1=beta_1, beta_2=beta_2, 
                                      epsilon=epsilon, 
                                      n_iter_no_change=n_iter_no_change, 
                                      max_fun=max_fun)
        
        self.name = 'Autoencoder'

    ###########################################################################
        
    def train_core (self):
        
        self.regressor.fit(self.X_train.values,
                           self.X_train.values.squeeze())
        
    ###########################################################################

    def map_from_X (self, X):
        
        return self.regressor.predict(X)
    
    ###########################################################################
    
    def set_hyperparameters(self, params_dict):     
        
        for key, value in params_dict.items():
            setattr(self.regressor, key, value)



import copy
import pandas as pd
import statsmodels.tsa.tsatools

###############################################################################

class PreProcess ():
    """
    Class used to encapsulate data preprocessing methods.
    
    Parameters
    ----------
            
        f_pp: list, optional
            List containing strings with names of methods to be used 
            in the preprocessing of the train data. The list of methods 
            is shown below.
        a_pp: dict, optional
            Dictionary containing the parameters to be provided
            to each function to perform preprocessing of the train data, in
            the format {'functionname__argname': argvalue, ...}
        is_Y: boolean, optional
            If the data being preprocessed is Y (that is, to be predicted).
        
    Methods:
    
    * Variable selection:
        remove_empty_variables();
        remove_frozen_variables()
        
    * Missing values imputation:
        ffill()
        remove_observations_with_nan();
        replace_nan_with_values()
        
    * Normalization:
        back_to_units();
        normalize()
        
    * Adding dynamics:
        apply_lag();
        add_moving_average()
        
    * Noise treatment:
        moving_average_filter()       
    
    """
            
    ###########################################################################
    
    def __init__(self, f_pp = None, a_pp = None, is_Y = False):
 
        self.is_Y = is_Y
        self.f_pp = f_pp        
        self._a_pp = a_pp     
        if self.f_pp is not None:
            self.params_per_func = {f: {} for f in f_pp}        

    ###########################################################################

    @property
    def a_pp(self):
        return self._a_pp

    ###########################################################################
    
    @a_pp.setter
    def a_pp(self, a_pp):

        self._a_pp = a_pp
        
        if self.f_pp is not None:
        
            self.params_per_func = {f: {} for f in self.f_pp}
            
            if a_pp is not None:
                                
                for pname, pval in a_pp.items():
                    func, param = pname.split('__',1)
                    self.params_per_func[func][param] = pval

    ###########################################################################
    
    def apply(self, df, train_or_test = 'train'):
        """
        Sequentially applies the preprocessing functions 
        defined during initialization.
    
        Parameters
        ----------
        df: pandas.DataFrame
            Data to be processed.
        train_or_test: string, optional
            Indicates which step the data corresponds to.
        Returns
        ----------                
        : pandas.DataFrame
           Processed data.
        """         
        
        df_processed = df
        
        for i in range(len(self.f_pp)):
            f = self.f_pp[i]
            df_processed = getattr(self, f)(df_processed,
                                            train_or_test, 
                                            **self.params_per_func[f])
                
        return df_processed
    
    ######################### 
    # VARIABLE SELECTION 
    #########################

    ###########################################################################

    def remove_empty_variables (self, df, train_or_test = 'train'):
        """
        Removes variables with no values.
    
        Parameters
        ----------
        df: pandas.DataFrame
            Data to be processed.
        train_or_test: string, optional
            Indicates which step the data corresponds to.
        Returns
        ----------                
        : pandas.DataFrame
           Processed data.
        """           
        if train_or_test == 'train':
            return df.dropna(axis=1, how='all')
        elif train_or_test == 'test':
            return df

    ###########################################################################

    def remove_frozen_variables (self, df, train_or_test = 'train',
                                 threshold = 1e-6): 
        """
        Removes variables whose variation falls below a given limit.
    
        Parameters
        ----------
        df: pandas.DataFrame
            Data to be processed.
        train_or_test: string, optional
            Indicates which step the data corresponds to.
        threshold: float, optional
            Variance limit to consider a variable as frozen.
        Returns
        ----------                
        : pandas.DataFrame
           Processed data.
        """                                              
        if not self.is_Y:
            if train_or_test == 'train':
                return df.loc[:, df.var(ddof=1) > threshold]
            elif train_or_test == 'test':
                return df
        else:
            return df
        
    ################################## 
    # MISSING VALUES IMPUTATION
    ##################################
        
    ###########################################################################
    
    def ffill_nan (self, df, train_or_test = 'train'):
        """
        Fills missing (NaN) values with the last valid value.
        Uses the next valid value if there is no last available.
    
        Parameters
        ----------
        df: pandas.DataFrame
            Data to be processed.
        train_or_test: string, optional
            Indicates which step the data corresponds to.
        Returns
        ----------                
        : pandas.DataFrame
           Processed data.
        """    
        return df.ffill().bfill()

    ###########################################################################
    
    def remove_observations_with_nan (self, df, train_or_test = 'train'):
        """
        Removes observations with missing data (NaN).
    
        Parameters
        ----------
        df: pandas.DataFrame
            Data to be processed.
        train_or_test: string, optional
            Indicates which step the data corresponds to.
        Returns
        ----------                
        : pandas.DataFrame
           Processed data.
        """    
        return df.dropna(axis=0, how='any')
    
    ###########################################################################

    def replace_nan_with_values (self, df, train_or_test = 'train', val = 0):
        """
        Replaces missing data (NaN) with a predefined value.

        Parameters
        ----------
        df: pandas.DataFrame
            Data to be processed.
        train_or_test: string, optional
            Indicates which step the data corresponds to.
        val: int or float
            Value to be used in the replacement.
        Returns
        ----------                
        : pandas.DataFrame
        Processed data.
        """    
                                    
        return df.fillna(val)

    ###############
    # NORMALIZATION
    ###############

    ###########################################################################

    def back_to_units (self, df):
        """
        Returns the variables to the original scale, 
        reverting effects of a normalization.

        Parameters
        ----------
        df: pandas.DataFrame
            Data to be processed.
        Returns
        ----------                
        : pandas.DataFrame
        Processed data.
        """    
        return df*self.SD + self.Mu
    
    ###########################################################################

    def normalize (self, df, train_or_test = 'train', mode = 'standard'):
        """
        Variable normalization.

        Parameters
        ----------
        df: pandas.DataFrame
            Data to be processed.
        train_or_test: string, optional
            Indicates which step the data corresponds to.
        mode: string, optional
            Type of normalization (standard, robust, m-robust or s-robust).
        Returns
        ----------                
        : pandas.DataFrame
        Processed data.
        """    
        if train_or_test == 'train':
            
            if mode == 'standard':
                self.Mu = df.mean()
                self.SD = df.std(ddof=1)
            elif mode == 'robust':
                self.Mu = df.median()
                self.SD = df.mad()               
            elif mode == 'm-robust':
                self.Mu = df.median()
                self.SD = df.std(ddof=1)
            elif mode == 's-robust':
                self.Mu = df.mean()
                self.SD = df.mad()               
            
            return (df - self.Mu)/self.SD
        
        elif train_or_test == 'test':
        
            return (df - self.Mu)/self.SD

    ##############################
    # ADDING DYNAMICS
    ##############################
        
    ###########################################################################

    def apply_lag (self, df, train_or_test = 'train', lag = 1):
        """
        Generation of time-delayed variables.

        Parameters
        ----------
        df: pandas.DataFrame
            Data to be processed.
        train_or_test: string, optional
            Indicates which step the data corresponds to.
        lag: int, optional
            Number of delays to be considered.
        Returns
        ----------                
        : pandas.DataFrame
        Processed data.
        """    
                        
        if self.is_Y:
            return df.iloc[lag:,:]
        else:    
            array_lagged = statsmodels.tsa.tsatools.lagmat(df, maxlag = lag, 
                                                           trim = "forward", 
                                                       original = 'in')[lag:,:]   
            new_columns = []
            for l in range(lag):
                new_columns.append(df.columns+' - lag '+str(l+1))
            columns_lagged = df.columns.append(new_columns)
            index_lagged = df.index[lag:]
            df_lagged = pd.DataFrame(array_lagged, index = index_lagged,
                                     columns = columns_lagged)
            
            return df_lagged  
        
    ###########################################################################

    def add_moving_average (self, df, train_or_test = 'train', WS = 10):
        """
        Adding variables filtered by moving average.
        Attention! Do not confuse with moving_average_filter, in which
        the original variables are not kept in the dataset.

        Parameters
        ----------
        df: pandas.DataFrame
            Data to be processed.
        train_or_test: string, optional
            Indicates which step the data corresponds to.
        WS: int, optional
            Window size of the filter.
        Returns
        ----------                
        : pandas.DataFrame
        Processed data.
        """    
        if self.is_Y:
            return df
                
        new_df = copy.deepcopy(df)
                
        for column in df:
            new_df[column+' MA'] = new_df[column].rolling(WS).mean()
        
        return new_df.drop(df.index[:WS])

    ##############################
    # NOISE TREATMENT
    ##############################

    ###########################################################################

    def moving_average_filter (self, df,  train_or_test = 'train', WS = 10):
        """
        Moving average noise filter.

        Parameters
        ----------
        df: pandas.DataFrame
            Data to be processed.
        train_or_test: string, optional
            Indicates which step the data corresponds to.
        WS: int, optional
            Window size of the filter.
        Returns
        ----------                
        : pandas.DataFrame
        Processed data.
        """    
        new_df = copy.deepcopy(df)
                
        for column in df:
            new_df[column] = new_df[column].rolling(WS).mean()
            
        if hasattr(df,'name'):
            new_df.name = df.name
                        
        return new_df.drop(df.index[:WS])


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Aug  4 20:03:43 2020

@author: afranio
"""

import setuptools
import os

with open('README.md') as f:
    README = f.read()

requirements = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    'requirements.txt'
)

if os.path.isfile(requirements):
    with open(requirements) as f:
        install_requires = f.read().splitlines()
else:
    install_requires = []

setuptools.setup(
    author="BibMon developers",
    author_email="cc-bibmon@petrobras.com.br",
    name='bibmon',
    description='Library with routines for data-driven process monitoring.',
    license='Apache 2.0',
    version='1.1.2',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/petrobras/bibmon',
    packages=setuptools.find_packages(include=['bibmon','bibmon.*']),
    include_package_data=True,
    package_data={
        'bibmon': ['real_process_data/*.csv','tennessee_eastman/*.dat'],
    },
    python_requires=">=3.12",
    install_requires=install_requires
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Sep  3 23:38:16 2020

@author: afranio
"""

import bibmon
import pandas as pd

def test_complete_analysis():
    
    # load data
    data = bibmon.load_real_data()
    data = data.apply(pd.to_numeric, errors='coerce')
    
    # preprocessing pipeline
    
    preproc_tr = ['remove_empty_variables',
                  'ffill_nan',
                  'remove_frozen_variables',
                  'normalize']
    
    preproc_ts = ['ffill_nan','normalize']
    
    # define training set
        
    (X_train, X_validation, 
     X_test, Y_train, 
     Y_validation, Y_test) = bibmon.train_val_test_split(data, 
                                            start_train = '2017-12-24T12:00', 
                                            end_train = '2018-01-01T00:00', 
                                           end_validation = '2018-01-02T00:00', 
                                            end_test = '2018-01-04T00:00',
                                            tags_Y = 'tag100')
                                                         
    # define the model
                                                         
    from sklearn.linear_model import LinearRegression
    reg = LinearRegression()
    
    model = bibmon.sklearnRegressor(reg)                                                          

    # define regression metrics
                                                         
    from sklearn.metrics import r2_score
    from sklearn.metrics import mean_absolute_error
    
    mtr = [r2_score, mean_absolute_error]
                           
    # complete analysis!
                              
    bibmon.complete_analysis(model, X_train, X_validation, X_test, 
                            Y_train, Y_validation, Y_test,
                            f_pp_train = preproc_tr,
                            f_pp_test = preproc_ts,
                            metrics = mtr, 
                            count_window_size = 3, count_limit = 2,
                            fault_start = '2018-01-02 06:00:00',
                            fault_end = '2018-01-02 09:00:00') 
    
    model.plot_importances()                                                                             


#%%
import bibmon
import pandas as pd
import numpy as np

#%%

def test_models_with_df_inputs():

    # load data
    data = bibmon.load_real_data()
    data = data.apply(pd.to_numeric, errors='coerce')
    
    # preprocessing pipeline
    f_pp_tr = ['remove_empty_variables',
               'ffill_nan',
               'remove_frozen_variables',
               'normalize']
    
    f_pp_ts = ['ffill_nan','normalize']
    
    # define training set
    (X_train, X_validation, 
     X_test, Y_train, 
     Y_validation, Y_test) = bibmon.train_val_test_split(data, 
                                           start_train = '2017-12-24T12:00', 
                                           end_train = '2018-01-01T00:00', 
                                           end_validation = '2018-01-02T00:00', 
                                           end_test = '2018-01-04T00:00',
                                           tags_Y = 'tag100')
    
    # running the unit tests
                                                             
    for attr in bibmon.__all__:             
        a = getattr(bibmon,attr)     
        if isinstance(a, type):         
            if a.__base__ == bibmon._generic_model.GenericModel:   
                if a == bibmon.sklearnRegressor:                 
                    from sklearn.linear_model import LinearRegression
                    m = a(LinearRegression())
                else:                    
                    m = a()        
                              
                # TRAINING
                    
                if m.has_Y:    
                    # X DF with one column
                    m.fit(X_train.loc[:,['tag102']], Y_train, 
                          f_pp = f_pp_tr,
                          delete_training_data = False)
                    
                    # X series with one column
                    m.fit(X_train.loc[:,['tag102']].T.squeeze(), Y_train, 
                          f_pp = f_pp_tr,
                          delete_training_data = False)
                    
                # X complete DF
                m.fit(X_train, Y_train, 
                      f_pp = f_pp_tr,
                      delete_training_data = False)
                                            
                dfs_train = [m.X_train, m.X_train_orig]
                series_train = [m.SPE_train]                

                if m.has_Y:
                    dfs_train.extend([m.Y_train_pred, m.Y_train_pred_orig])
                else:
                    dfs_train.extend([m.X_train_pred, m.X_train_pred_orig])
                       
                if m.has_Y:
                    dfs_train.extend([m.Y_train, m.Y_train_orig])
                
                for df in dfs_train:
                    assert(isinstance(df, pd.DataFrame))
                    
                for s in series_train:
                    assert(isinstance(s, pd.Series))
                    
                assert(isinstance(m.limSPE, float))

                # PREDICTION 1 - VALIDATION AND REDEFINITION OF LIMIT
                
                m.predict(X_validation, Y_validation, redefine_limit = True,
                          f_pp = f_pp_ts)
                
                dfs_val = [m.X_test, m.X_test_orig]
                series_val = [m.SPE_test]                

                if m.has_Y:
                    dfs_val.extend([m.Y_test_pred, m.Y_test_pred_orig])
                else:
                    dfs_val.extend([m.X_test_pred, m.X_test_pred_orig])
                       
                if m.has_Y:
                    dfs_val.extend([m.Y_test, m.Y_test_orig])
                
                for df in dfs_val:
                    assert(isinstance(df, pd.DataFrame))
                    
                for s in series_val:
                    assert(isinstance(s, pd.Series))
                
                # PREDICTION 2 - TEST WITH WINDOWS FOR ALARMS
                    
                m.predict(X_test, Y_test,                           
                          count_window_size = 10,
                          redefine_limit = False, count_limit = 5, 
                          f_pp = f_pp_ts)        
                                
                series_alarms = [m.alarmOutlier, m.alarmCount]

                for s in series_alarms:
                    assert(isinstance(s, pd.Series))
                    
                # PREDICTION 3 - TEST WITH A POINT (SERIES AND DF)

                f_pp_ts =['replace_nan_with_values']
                a_pp_ts = {'replace_nan_with_values__val': X_train.median()}
        
                # SERIES
                m.predict(X_test.iloc[0:1], Y_test.iloc[0:1],
                          f_pp = f_pp_ts,
                          a_pp = a_pp_ts)
                
                # DF
                m.predict(X_test.iloc[0:1], Y_test.iloc[0:1],
                          f_pp = f_pp_ts,
                          a_pp = a_pp_ts)               

#%%
                
def test_models_with_np_array_inputs():

    # load data
    train_df, test_df = bibmon.load_tennessee_eastman(train_id = 0, 
                                                      test_id = 1)
    
    X_train = train_df.drop('XMEAS(35)',axis=1)
    Y_train = train_df['XMEAS(35)']

    X_test = test_df.drop('XMEAS(35)',axis=1)
    Y_test = test_df['XMEAS(35)']

    # preprocessing pipeline    
    f_pp = ['normalize']
    
    for attr in bibmon.__all__:             
        a = getattr(bibmon,attr)     
        if isinstance(a, type):         
            if a.__base__ == bibmon._generic_model.GenericModel:   
                if a == bibmon.sklearnRegressor:                 
                    from sklearn.linear_model import LinearRegression
                    m = a(LinearRegression())
                else:                    
                    m = a()        
                              
                # TRAINING

                if m.has_Y:    
                    # X with one column
                    m.fit(np.array(X_train.iloc[:,0]), np.array(Y_train), 
                          f_pp = f_pp,
                          delete_training_data = False)
                    
                # complete X
                m.fit(np.array(X_train), np.array(Y_train), 
                      f_pp = f_pp,
                      delete_training_data = False)
    
                dfs_train = [m.X_train, m.X_train_orig]
                series_train = [m.SPE_train]                

                if m.has_Y:
                    dfs_train.extend([m.Y_train_pred, m.Y_train_pred_orig])
                else:
                    dfs_train.extend([m.X_train_pred, m.X_train_pred_orig])
                       
                if m.has_Y:
                    dfs_train.extend([m.Y_train, m.Y_train_orig])
                
                for df in dfs_train:
                    assert(isinstance(df, pd.DataFrame))
                    
                for s in series_train:
                    assert(isinstance(s, pd.Series))
                    
                assert(isinstance(m.limSPE, float))  
                
                if isinstance(m, bibmon.PCA):
                    m.plot_cumulative_variance()
                
                # TEST
                
                m.predict(np.array(X_test), np.array(Y_test), 
                          f_pp = f_pp)
                
                dfs_val = [m.X_test, m.X_test_orig]
                series_val = [m.SPE_test]                

                if m.has_Y:
                    dfs_val.extend([m.Y_test_pred, m.Y_test_pred_orig])
                else:
                    dfs_val.extend([m.X_test_pred, m.X_test_pred_orig])
                       
                if m.has_Y:
                    dfs_val.extend([m.Y_test, m.Y_test_orig])
                
                for df in dfs_val:
                    assert(isinstance(df, pd.DataFrame))
                    
                for s in series_val:
                    assert(isinstance(s, pd.Series))
                    
                # PREDICTION 3 - TEST WITH A POINT (SERIES AND DF)

                f_pp_ts =['normalize','replace_nan_with_values']
                a_pp_ts = {'replace_nan_with_values__val': X_train.median()}
        
                # SERIES
                m.predict(np.array(X_test.iloc[0]), np.array(Y_test.iloc[0]),
                          f_pp = f_pp_ts,
                          a_pp = a_pp_ts)
                
                # DF
                m.predict(np.array(X_test.iloc[0:1]), 
                          np.array(Y_test.iloc[0:1]),
                          f_pp = f_pp_ts,
                          a_pp = a_pp_ts)           
                